# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any
import functools

from mautrix import __optional_imports__

if __optional_imports__:
    from typing import Awaitable, Callable, ParamSpec

    Param = ParamSpec("Param")
    Func = Callable[Param, Awaitable[Any]]


def async_getter_lock(fn: Func) -> Func:
    """
    A utility decorator for locking async getters that have caches
    (preventing race conditions between cache check and e.g. async database actions).

    The class must have an ```_async_get_locks`` defaultdict that contains :class:`asyncio.Lock`s
    (see example for exact definition). Non-cache-affecting arguments should be only passed as
    keyword args.

    Args:
        fn: The function to decorate.

    Returns:
        The decorated function.

    Examples:
        >>> import asyncio
        >>> from collections import defaultdict
        >>> class User:
        ...   _async_get_locks: dict[Any, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())
        ...   db: Any
        ...   cache: dict[str, User]
        ...   @classmethod
        ...   @async_getter_lock
        ...   async def get(cls, id: str, *, create: bool = False) -> User | None:
        ...     try:
        ...       return cls.cache[id]
        ...     except KeyError:
        ...       pass
        ...     user = await cls.db.fetch_user(id)
        ...     if user:
        ...       return user
        ...     elif create:
        ...       return await cls.db.create_user(id)
        ...     return None
    """

    @functools.wraps(fn)
    async def wrapper(cls, *args, **kwargs) -> Any:
        async with cls._async_get_locks[args]:
            return await fn(cls, *args, **kwargs)

    return wrapper



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Callable, Dict, Iterator, NewType, Optional, Tuple, Type, TypeVar, Union
from uuid import UUID
import copy
import logging

import attr

from ..primitive import JSON
from .obj import Lst, Obj
from .serializable import (
    AbstractSerializable,
    Serializable,
    SerializableSubtype,
    SerializerError,
    UnknownSerializationError,
)

T = TypeVar("T")
T2 = TypeVar("T2")

Serializer = NewType("Serializer", Callable[[T], JSON])
Deserializer = NewType("Deserializer", Callable[[JSON], T])
serializer_map: Dict[Type[T], Serializer] = {
    UUID: str,
}
deserializer_map: Dict[Type[T], Deserializer] = {
    UUID: UUID,
}

META_JSON = "json"
META_FLATTEN = "flatten"
META_HIDDEN = "hidden"
META_IGNORE_ERRORS = "ignore_errors"
META_OMIT_EMPTY = "omitempty"
META_OMIT_DEFAULT = "omitdefault"

log = logging.getLogger("mau.attrs")


def field(
    default: Any = attr.NOTHING,
    factory: Optional[Callable[[], Any]] = None,
    json: Optional[str] = None,
    flatten: bool = False,
    hidden: bool = False,
    ignore_errors: bool = False,
    omit_empty: bool = True,
    omit_default: bool = False,
    metadata: Optional[Dict[str, Any]] = None,
    **kwargs,
):
    """
    A wrapper around :meth:`attr.ib` to conveniently add SerializableAttrs metadata fields.

    Args:
        default: Same as attr.ib, the default value for the field.
        factory: Same as attr.ib, a factory function that creates the default value.
        json: The JSON key used for de/serializing the object.
        flatten: Set to flatten subfields inside this field to be a part of the parent object in
            serialized objects. When deserializing, the input data will be deserialized into both
            the parent and child fields, so the classes should ignore unknown keys.
        hidden: Set to always omit the key from serialized objects.
        ignore_errors: Set to ignore type errors while deserializing.
        omit_empty: Set to omit the key from serialized objects if the value is ``None``.
        omit_default: Set to omit the key from serialized objects if the value is equal to the
            default.
        metadata: Additional metadata for attr.ib.
        **kwargs: Additional keyword arguments for attr.ib.

    Returns:
        The decorator function returned by attr.ib.

    Examples:
        >>> from attr import dataclass
        >>> from mautrix.types import SerializableAttrs, field
        >>> @dataclass
        ... class SomeData(SerializableAttrs):
        ...     my_field: str = field(json="com.example.namespaced_field", default="hi")
        ...
        >>> SomeData().serialize()
        {'com.example.namespaced_field': 'hi'}
        >>> SomeData.deserialize({"com.example.namespaced_field": "hmm"})
        SomeData(my_field='hmm')
    """
    custom_meta = {
        META_JSON: json,
        META_FLATTEN: flatten,
        META_HIDDEN: hidden,
        META_IGNORE_ERRORS: ignore_errors,
        META_OMIT_EMPTY: omit_empty,
        META_OMIT_DEFAULT: omit_default,
    }
    metadata = metadata or {}
    metadata.update({k: v for k, v in custom_meta.items() if v is not None})
    return attr.ib(default=default, factory=factory, metadata=metadata, **kwargs)


def serializer(elem_type: Type[T]) -> Callable[[Serializer], Serializer]:
    """
    Define a custom serialization function for the given type.

    Args:
        elem_type: The type to define the serializer for.

    Returns:
        Decorator for the function. The decorator will simply add the function to a map of
        deserializers and return the function.

    Examples:
        >>> from datetime import datetime
        >>> from mautrix.types import serializer, JSON
        >>> @serializer(datetime)
        ... def serialize_datetime(dt: datetime) -> JSON:
        ...     return dt.timestamp()
    """

    def decorator(func: Serializer) -> Serializer:
        serializer_map[elem_type] = func
        return func

    return decorator


def deserializer(elem_type: Type[T]) -> Callable[[Deserializer], Deserializer]:
    """
    Define a custom deserialization function for a given type hint.

    Args:
        elem_type: The type hint to define the deserializer for.

    Returns:
        Decorator for the function. The decorator will simply add the function to a map of
        deserializers and return the function.

    Examples:
        >>> from datetime import datetime
        >>> from mautrix.types import deserializer, JSON
        >>> @deserializer(datetime)
        ... def deserialize_datetime(data: JSON) -> datetime:
        ...     return datetime.fromtimestamp(data)
    """

    def decorator(func: Deserializer) -> Deserializer:
        deserializer_map[elem_type] = func
        return func

    return decorator


def _fields(attrs_type: Type[T], only_if_flatten: bool = None) -> Iterator[Tuple[str, Type[T2]]]:
    for field in attr.fields(attrs_type):
        if field.metadata.get(META_HIDDEN, False):
            continue
        if only_if_flatten is None or field.metadata.get(META_FLATTEN, False) == only_if_flatten:
            yield field.metadata.get(META_JSON, field.name), field


immutable = int, str, float, bool, type(None)


def _safe_default(val: T) -> T:
    if isinstance(val, immutable):
        return val
    elif val is attr.NOTHING:
        return None
    elif isinstance(val, attr.Factory):
        if val.takes_self:
            # TODO implement?
            return None
        else:
            return val.factory()
    return copy.copy(val)


def _dict_to_attrs(
    attrs_type: Type[T], data: JSON, default: Optional[T] = None, default_if_empty: bool = False
) -> T:
    data = data or {}
    unrecognized = {}
    new_items = {
        field_meta.name.lstrip("_"): _try_deserialize(field_meta, data)
        for _, field_meta in _fields(attrs_type, only_if_flatten=True)
    }
    fields = dict(_fields(attrs_type, only_if_flatten=False))
    for key, value in data.items():
        try:
            field_meta = fields[key]
        except KeyError:
            unrecognized[key] = value
            continue
        name = field_meta.name.lstrip("_")
        try:
            new_items[name] = _try_deserialize(field_meta, value)
        except UnknownSerializationError as e:
            raise SerializerError(
                f"Failed to deserialize {value} into key {name} of {attrs_type.__name__}"
            ) from e
        except SerializerError:
            raise
        except Exception as e:
            raise SerializerError(
                f"Failed to deserialize {value} into key {name} of {attrs_type.__name__}"
            ) from e
    if len(new_items) == 0 and default_if_empty and default is not attr.NOTHING:
        return _safe_default(default)
    try:
        obj = attrs_type(**new_items)
    except TypeError as e:
        for key, field_meta in _fields(attrs_type):
            if field_meta.default is attr.NOTHING and key not in new_items:
                log.debug("Failed to deserialize %s into %s", data, attrs_type.__name__)
                json_key = field_meta.metadata.get(META_JSON, key)
                raise SerializerError(
                    f"Missing value for required key {json_key} in {attrs_type.__name__}"
                ) from e
        raise UnknownSerializationError() from e
    if len(unrecognized) > 0:
        obj.unrecognized_ = unrecognized
    return obj


def _try_deserialize(field, value: JSON) -> T:
    try:
        return _deserialize(field.type, value, field.default)
    except SerializerError:
        if not field.metadata.get(META_IGNORE_ERRORS, False):
            raise
    except (TypeError, ValueError, KeyError) as e:
        if not field.metadata.get(META_IGNORE_ERRORS, False):
            raise UnknownSerializationError() from e


def _has_custom_deserializer(cls) -> bool:
    return issubclass(cls, Serializable) and getattr(cls.deserialize, "__func__") != getattr(
        SerializableAttrs.deserialize, "__func__"
    )


def _deserialize(cls: Type[T], value: JSON, default: Optional[T] = None) -> T:
    if value is None:
        return _safe_default(default)

    try:
        deser = deserializer_map[cls]
    except KeyError:
        pass
    else:
        return deser(value)
    supertype = getattr(cls, "__supertype__", None)
    if supertype:
        cls = supertype
        try:
            deser = deserializer_map[supertype]
        except KeyError:
            pass
        else:
            return deser(value)

    if attr.has(cls):
        if _has_custom_deserializer(cls):
            return cls.deserialize(value)
        return _dict_to_attrs(cls, value, default, default_if_empty=True)
    elif cls == Any or cls == JSON:
        return value
    elif isinstance(cls, type) and issubclass(cls, Serializable):
        return cls.deserialize(value)

    type_class = getattr(cls, "__origin__", None)
    args = getattr(cls, "__args__", None)
    if type_class is Union:
        if len(args) == 2 and isinstance(None, args[1]):
            return _deserialize(args[0], value, default)
    elif type_class == list:
        (item_cls,) = args
        return [_deserialize(item_cls, item) for item in value]
    elif type_class == set:
        (item_cls,) = args
        return {_deserialize(item_cls, item) for item in value}
    elif type_class == dict:
        key_cls, val_cls = args
        return {
            _deserialize(key_cls, key): _deserialize(val_cls, item) for key, item in value.items()
        }

    if isinstance(value, list):
        return Lst(value)
    elif isinstance(value, dict):
        return Obj(**value)
    return value


def _actual_type(cls: Type[T]) -> Type[T]:
    if cls is None:
        return cls
    if getattr(cls, "__origin__", None) is Union:
        if len(cls.__args__) == 2 and isinstance(None, cls.__args__[1]):
            return cls.__args__[0]
    return cls


def _get_serializer(cls: Type[T]) -> Serializer:
    return serializer_map.get(_actual_type(cls), _serialize)


def _serialize_attrs_field(data: T, field: T2) -> JSON:
    field_val = getattr(data, field.name)
    if field_val is None:
        if not field.metadata.get(META_OMIT_EMPTY, True):
            if field.default is not attr.NOTHING:
                field_val = _safe_default(field.default)
        else:
            return attr.NOTHING

    if field.metadata.get(META_OMIT_DEFAULT, False) and field_val == field.default:
        return attr.NOTHING

    return _get_serializer(field.type)(field_val)


def _attrs_to_dict(data: T) -> JSON:
    new_dict = {}
    for json_name, field in _fields(data.__class__):
        if not json_name:
            continue
        serialized = _serialize_attrs_field(data, field)
        if serialized is not attr.NOTHING:
            if field.metadata.get(META_FLATTEN, False) and isinstance(serialized, dict):
                new_dict.update(serialized)
            else:
                new_dict[json_name] = serialized
    try:
        new_dict.update(data.unrecognized_)
    except (AttributeError, TypeError):
        pass
    return new_dict


def _serialize(val: Any) -> JSON:
    if isinstance(val, Serializable):
        return val.serialize()
    elif isinstance(val, (tuple, list, set)):
        return [_serialize(subval) for subval in val]
    elif isinstance(val, dict):
        return {_serialize(subkey): _serialize(subval) for subkey, subval in val.items()}
    elif attr.has(val.__class__):
        return _attrs_to_dict(val)
    return val


class SerializableAttrs(AbstractSerializable):
    """
    An abstract :class:`Serializable` that assumes the subclass is an attrs dataclass.

    Examples:
        >>> from attr import dataclass
        >>> from mautrix.types import SerializableAttrs
        >>> @dataclass
        ... class Foo(SerializableAttrs):
        ...     index: int
        ...     field: Optional[str] = None
    """

    unrecognized_: Dict[str, JSON]

    def __init__(self):
        self.unrecognized_ = {}

    @classmethod
    def deserialize(cls: Type[SerializableSubtype], data: JSON) -> SerializableSubtype:
        return _dict_to_attrs(cls, data)

    def serialize(self) -> JSON:
        return _attrs_to_dict(self)

    def get(self, item: str, default: Any = None) -> Any:
        try:
            return self[item]
        except KeyError:
            return default

    def __contains__(self, item: str) -> bool:
        return hasattr(self, item) or item in getattr(self, "unrecognized_", {})

    def __getitem__(self, item: str) -> Any:
        try:
            return getattr(self, item)
        except AttributeError:
            try:
                return self.unrecognized_[item]
            except AttributeError:
                self.unrecognized_ = {}
                raise KeyError(item)

    def __setitem__(self, item: str, value: Any) -> None:
        if hasattr(self, item):
            setattr(self, item, value)
        else:
            try:
                self.unrecognized_[item] = value
            except AttributeError:
                self.unrecognized_ = {item: value}



# Copyright © 2019 Damir Jelić <poljar@termina.org.uk> (under the Apache 2.0 license)
# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pytest
import unpaddedbase64

from mautrix.errors import DecryptionError

from .attachments import decrypt_attachment, encrypt_attachment, inplace_encrypt_attachment

try:
    from Crypto import Random
except ImportError:
    from Cryptodome import Random


def test_encrypt():
    data = b"Test bytes"

    cyphertext, keys = encrypt_attachment(data)

    plaintext = decrypt_attachment(cyphertext, keys.key.key, keys.hashes["sha256"], keys.iv)

    assert data == plaintext


def test_inplace_encrypt():
    orig_data = b"Test bytes"
    data = bytearray(orig_data)

    keys = inplace_encrypt_attachment(data)

    assert data != orig_data

    decrypt_attachment(data, keys.key.key, keys.hashes["sha256"], keys.iv, inplace=True)

    assert data == orig_data


def test_hash_verification():
    data = b"Test bytes"

    cyphertext, keys = encrypt_attachment(data)

    with pytest.raises(DecryptionError):
        decrypt_attachment(cyphertext, keys.key.key, "Fake hash", keys.iv)


def test_invalid_key():
    data = b"Test bytes"

    cyphertext, keys = encrypt_attachment(data)

    with pytest.raises(DecryptionError):
        decrypt_attachment(cyphertext, "Fake key", keys.hashes["sha256"], keys.iv)


def test_invalid_iv():
    data = b"Test bytes"

    cyphertext, keys = encrypt_attachment(data)

    with pytest.raises(DecryptionError):
        decrypt_attachment(cyphertext, keys.key.key, keys.hashes["sha256"], "Fake iv")


def test_short_key():
    data = b"Test bytes"

    cyphertext, keys = encrypt_attachment(data)

    with pytest.raises(DecryptionError):
        decrypt_attachment(
            cyphertext,
            unpaddedbase64.encode_base64(b"Fake key", urlsafe=True),
            keys["hashes"]["sha256"],
            keys["iv"],
        )


def test_short_iv():
    data = b"Test bytes"

    cyphertext, keys = encrypt_attachment(data)

    with pytest.raises(DecryptionError):
        decrypt_attachment(
            cyphertext,
            keys.key.key,
            keys.hashes["sha256"],
            unpaddedbase64.encode_base64(b"F" + b"\x00" * 8),
        )


def test_fake_key():
    data = b"Test bytes"

    cyphertext, keys = encrypt_attachment(data)

    fake_key = Random.new().read(32)

    plaintext = decrypt_attachment(
        cyphertext,
        unpaddedbase64.encode_base64(fake_key, urlsafe=True),
        keys["hashes"]["sha256"],
        keys["iv"],
    )
    assert plaintext != data



# This script generates the __all__ arrays for types/__init__.py and errors/__init__.py
# to avoid having to manually add both the import and the __all__ entry.
# See https://github.com/mautrix/python/issues/90 for why __all__ is needed at all.
from pathlib import Path
import ast

import black

root_module = Path(__file__).parent

black_cfg = black.parse_pyproject_toml(str(root_module.parent / "pyproject.toml"))
black_mode = black.Mode(
    target_versions={black.TargetVersion[ver.upper()] for ver in black_cfg["target_version"]},
    line_length=black_cfg["line_length"],
)


def add_imports_to_all(dir: str) -> None:
    init_file = root_module / dir / "__init__.py"
    with open(init_file) as f:
        init_ast = ast.parse(f.read(), filename=f"mautrix/{dir}/__init__.py")

    imports: list[str] = []
    all_node: ast.List | None = None

    for node in ast.iter_child_nodes(init_ast):
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            imports += (name.name for name in node.names)
        elif isinstance(node, ast.Assign) and isinstance(node.value, ast.List):
            target = node.targets[0]
            if len(node.targets) == 1 and isinstance(target, ast.Name) and target.id == "__all__":
                all_node = node.value

    all_node.elts = [ast.Constant(name) for name in imports]

    with open(init_file, "w") as f:
        f.write(black.format_str(ast.unparse(init_ast), mode=black_mode))


add_imports_to_all("types")
add_imports_to_all("errors")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .account_data import (
    AccountDataEvent,
    AccountDataEventContent,
    RoomTagAccountDataEventContent,
    RoomTagInfo,
)
from .base import BaseEvent, BaseRoomEvent, BaseUnsigned, GenericEvent
from .batch import BatchSendEvent, BatchSendStateEvent
from .beeper import (
    BeeperMessageStatusEvent,
    BeeperMessageStatusEventContent,
    MessageStatus,
    MessageStatusReason,
)
from .encrypted import (
    EncryptedEvent,
    EncryptedEventContent,
    EncryptedMegolmEventContent,
    EncryptedOlmEventContent,
    EncryptionAlgorithm,
    EncryptionKeyAlgorithm,
    KeyID,
    OlmCiphertext,
    OlmMsgType,
)
from .ephemeral import (
    EphemeralEvent,
    PresenceEvent,
    PresenceEventContent,
    PresenceState,
    ReceiptEvent,
    ReceiptEventContent,
    ReceiptType,
    SingleReceiptEventContent,
    TypingEvent,
    TypingEventContent,
)
from .generic import Event, EventContent
from .message import (
    AudioInfo,
    BaseFileInfo,
    BaseMessageEventContent,
    BaseMessageEventContentFuncs,
    EncryptedFile,
    FileInfo,
    Format,
    ImageInfo,
    InReplyTo,
    JSONWebKey,
    LocationInfo,
    LocationMessageEventContent,
    MediaInfo,
    MediaMessageEventContent,
    MessageEvent,
    MessageEventContent,
    MessageType,
    MessageUnsigned,
    RelatesTo,
    RelationType,
    TextMessageEventContent,
    ThumbnailInfo,
    VideoInfo,
)
from .reaction import ReactionEvent, ReactionEventContent
from .redaction import RedactionEvent, RedactionEventContent
from .state import (
    CanonicalAliasStateEventContent,
    JoinRule,
    JoinRulesStateEventContent,
    Membership,
    MemberStateEventContent,
    PowerLevelStateEventContent,
    RoomAvatarStateEventContent,
    RoomCreateStateEventContent,
    RoomEncryptionStateEventContent,
    RoomNameStateEventContent,
    RoomPinnedEventsStateEventContent,
    RoomPredecessor,
    RoomTombstoneStateEventContent,
    RoomTopicStateEventContent,
    RoomType,
    SpaceChildStateEventContent,
    SpaceParentStateEventContent,
    StateEvent,
    StateEventContent,
    StateUnsigned,
    StrippedStateEvent,
)
from .to_device import (
    ASToDeviceEvent,
    ForwardedRoomKeyEventContent,
    KeyRequestAction,
    RequestedKeyInfo,
    RoomKeyEventContent,
    RoomKeyRequestEventContent,
    RoomKeyWithheldCode,
    RoomKeyWithheldEventContent,
    ToDeviceEvent,
    ToDeviceEventContent,
)
from .type import EventType
from .voip import (
    CallAnswerEventContent,
    CallCandidate,
    CallCandidatesEventContent,
    CallData,
    CallDataType,
    CallEvent,
    CallEventContent,
    CallHangupEventContent,
    CallHangupReason,
    CallInviteEventContent,
    CallNegotiateEventContent,
    CallRejectEventContent,
    CallSelectAnswerEventContent,
)



# Copyright (c) 2023 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import AsyncGenerator, Union
import logging

import aiohttp

AsyncBody = AsyncGenerator[Union[bytes, bytearray, memoryview], None]


async def async_iter_bytes(data: bytearray | bytes, chunk_size: int = 1024**2) -> AsyncBody:
    """
    Return memory views into a byte array in chunks. This is used to prevent aiohttp from copying
    the entire request body.

    Args:
        data: The underlying data to iterate through.
        chunk_size: How big each returned chunk should be.

    Returns:
        An async generator that yields the given data in chunks.
    """
    with memoryview(data) as mv:
        for i in range(0, len(data), chunk_size):
            yield mv[i : i + chunk_size]


class FileTooLargeError(Exception):
    def __init__(self, max_size: int) -> None:
        super().__init__(f"File size larger than maximum ({max_size / 1024 / 1024} MiB)")


_default_dl_log = logging.getLogger("mau.util.download")


async def read_response_chunks(
    resp: aiohttp.ClientResponse, max_size: int, log: logging.Logger = _default_dl_log
) -> bytearray:
    """
    Read the body from an aiohttp response in chunks into a mutable bytearray.

    Args:
        resp: The aiohttp response object to read the body from.
        max_size: The maximum size to read. FileTooLargeError will be raised if the Content-Length
            is higher than this, or if the body exceeds this size during reading.
        log: A logger for logging download status.

    Returns:
        The body data as a byte array.

    Raises:
        FileTooLargeError: if the body is larger than the provided max_size.
    """
    content_length = int(resp.headers.get("Content-Length", "0"))
    if 0 < max_size < content_length:
        raise FileTooLargeError(max_size)
    size_str = "unknown length" if content_length == 0 else f"{content_length} bytes"
    log.info(f"Reading file download response with {size_str} (max: {max_size})")
    data = bytearray(content_length)
    mv = memoryview(data) if content_length > 0 else None
    read_size = 0
    max_size += 1
    while True:
        block = await resp.content.readany()
        if not block:
            break
        max_size -= len(block)
        if max_size <= 0:
            raise FileTooLargeError(max_size)
        if len(data) >= read_size + len(block):
            mv[read_size : read_size + len(block)] = block
        elif len(data) > read_size:
            log.warning("File being downloaded is bigger than expected")
            mv[read_size:] = block[: len(data) - read_size]
            mv.release()
            mv = None
            data.extend(block[len(data) - read_size :])
        else:
            if mv is not None:
                mv.release()
                mv = None
            data.extend(block)
        read_size += len(block)
    if mv is not None:
        mv.release()
    log.info(f"Successfully read {read_size} bytes of file download response")
    return data


__all__ = ["AsyncBody", "FileTooLargeError", "async_iter_bytes", "async_read_bytes"]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from mautrix.errors import DeviceValidationError
from mautrix.types import (
    CrossSigner,
    CrossSigningKeys,
    CrossSigningUsage,
    DeviceID,
    DeviceIdentity,
    DeviceKeys,
    EncryptionKeyAlgorithm,
    IdentityKey,
    KeyID,
    QueryKeysResponse,
    SigningKey,
    SyncToken,
    TrustState,
    UserID,
)

from .base import BaseOlmMachine, verify_signature_json


class DeviceListMachine(BaseOlmMachine):
    async def _fetch_keys(
        self, users: list[UserID], since: SyncToken = "", include_untracked: bool = False
    ) -> dict[UserID, dict[DeviceID, DeviceIdentity]]:
        if not include_untracked:
            users = await self.crypto_store.filter_tracked_users(users)
        if len(users) == 0:
            return {}
        users = set(users)

        self.log.trace(f"Querying keys for {users}")
        resp = await self.client.query_keys(users, token=since)
        missing_users = users.copy()

        for server, err in resp.failures.items():
            self.log.warning(f"Query keys failure for {server}: {err}")

        data = {}
        for user_id, devices in resp.device_keys.items():
            missing_users.remove(user_id)

            new_devices = {}
            existing_devices = (await self.crypto_store.get_devices(user_id)) or {}

            self.log.trace(
                f"Updating devices for {user_id}, got {len(devices)}, "
                f"have {len(existing_devices)} in store"
            )
            changed = False
            ssks = resp.self_signing_keys.get(user_id)
            ssk = ssks.first_ed25519_key if ssks else None
            for device_id, device_keys in devices.items():
                try:
                    existing = existing_devices[device_id]
                except KeyError:
                    existing = None
                    changed = True
                self.log.trace(f"Validating device {device_keys} of {user_id}")
                try:
                    new_device = await self._validate_device(
                        user_id, device_id, device_keys, existing
                    )
                except DeviceValidationError as e:
                    self.log.warning(f"Failed to validate device {device_id} of {user_id}: {e}")
                else:
                    if new_device:
                        new_devices[device_id] = new_device
                        await self._store_device_self_signatures(device_keys, ssk)
            self.log.debug(
                f"Storing new device list for {user_id} containing {len(new_devices)} devices"
            )
            await self.crypto_store.put_devices(user_id, new_devices)
            data[user_id] = new_devices

            if changed or len(new_devices) != len(existing_devices):
                if self.delete_keys_on_device_delete:
                    for device_id in existing_devices.keys() - new_devices.keys():
                        device = existing_devices[device_id]
                        removed_ids = await self.crypto_store.redact_group_sessions(
                            room_id=None, sender_key=device.identity_key, reason="device removed"
                        )
                        self.log.info(
                            "Redacted megolm sessions sent by removed device "
                            f"{device.user_id}/{device.device_id}: {removed_ids}"
                        )
                await self.on_devices_changed(user_id)

        for user_id in missing_users:
            self.log.warning(f"Didn't get any devices for user {user_id}")

        for user_id in users:
            await self._store_cross_signing_keys(resp, user_id)

        return data

    async def _store_device_self_signatures(
        self, device_keys: DeviceKeys, self_signing_key: SigningKey | None
    ) -> None:
        device_desc = f"Device {device_keys.user_id}/{device_keys.device_id}"
        try:
            self_signatures = device_keys.signatures[device_keys.user_id].copy()
        except KeyError:
            self.log.warning(f"{device_desc} doesn't have any signatures from the user")
            return
        if len(device_keys.signatures) > 1:
            self.log.debug(
                f"{device_desc} has signatures from other users (%s)",
                set(device_keys.signatures.keys()) - {device_keys.user_id},
            )

        device_self_sig = self_signatures.pop(
            KeyID(EncryptionKeyAlgorithm.ED25519, device_keys.device_id)
        )
        target = CrossSigner(device_keys.user_id, device_keys.ed25519)
        # This one is already validated by _validate_device
        await self.crypto_store.put_signature(target, target, device_self_sig)

        try:
            cs_self_sig = self_signatures.pop(
                KeyID(EncryptionKeyAlgorithm.ED25519, self_signing_key)
            )
        except KeyError:
            self.log.warning(f"{device_desc} isn't cross-signed")
        else:
            is_valid_self_sig = verify_signature_json(
                device_keys.serialize(), device_keys.user_id, self_signing_key, self_signing_key
            )
            if is_valid_self_sig:
                signer = CrossSigner(device_keys.user_id, self_signing_key)
                await self.crypto_store.put_signature(target, signer, cs_self_sig)
            else:
                self.log.warning(f"{device_desc} doesn't have a valid cross-signing signature")

        if len(self_signatures) > 0:
            self.log.debug(
                f"{device_desc} has signatures from unexpected keys (%s)",
                set(self_signatures.keys()),
            )

    async def _store_cross_signing_keys(self, resp: QueryKeysResponse, user_id: UserID) -> None:
        new_keys: dict[CrossSigningUsage, CrossSigningKeys] = {}
        try:
            master = new_keys[CrossSigningUsage.MASTER] = resp.master_keys[user_id]
        except KeyError:
            self.log.debug(f"Didn't get a cross-signing master key for {user_id}")
            return
        try:
            new_keys[CrossSigningUsage.SELF] = resp.self_signing_keys[user_id]
        except KeyError:
            self.log.debug(f"Didn't get a cross-signing self-signing key for {user_id}")
            return
        try:
            new_keys[CrossSigningUsage.USER] = resp.user_signing_keys[user_id]
        except KeyError:
            pass
        current_keys = await self.crypto_store.get_cross_signing_keys(user_id)
        for usage, key in current_keys.items():
            if usage in new_keys and key.key != new_keys[usage].first_ed25519_key:
                num = await self.crypto_store.drop_signatures_by_key(CrossSigner(user_id, key.key))
                if num >= 0:
                    self.log.debug(
                        f"Dropped {num} signatures made by key {user_id}/{key.key} ({usage})"
                        " as it has been replaced"
                    )
        for usage, key in new_keys.items():
            actual_key = key.first_ed25519_key
            self.log.debug(f"Storing cross-signing key for {user_id}: {actual_key} (type {usage})")
            await self.crypto_store.put_cross_signing_key(user_id, usage, actual_key)

            if usage != CrossSigningUsage.MASTER and (
                KeyID(EncryptionKeyAlgorithm.ED25519, master.first_ed25519_key)
                not in key.signatures[user_id]
            ):
                self.log.warning(
                    f"Cross-signing key {user_id}/{actual_key}/{usage}"
                    " doesn't seem to have a signature from the master key"
                )

            for signer_user_id, signatures in key.signatures.items():
                for key_id, signature in signatures.items():
                    signing_key = SigningKey(key_id.key_id)
                    if signer_user_id == user_id:
                        try:
                            device = resp.device_keys[signer_user_id][DeviceID(key_id.key_id)]
                            signing_key = device.ed25519
                        except KeyError:
                            pass
                    if len(signing_key) != 43:
                        self.log.debug(
                            f"Cross-signing key {user_id}/{actual_key} has a signature from "
                            f"an unknown key {key_id}"
                        )
                        continue
                    signing_key_log = signing_key
                    if signing_key != key_id.key_id:
                        signing_key_log = f"{signing_key} ({key_id})"
                    self.log.debug(
                        f"Verifying cross-signing key {user_id}/{actual_key} "
                        f"with key {signer_user_id}/{signing_key_log}"
                    )
                    is_valid_sig = verify_signature_json(
                        key.serialize(), signer_user_id, key_id.key_id, signing_key
                    )
                    if is_valid_sig:
                        self.log.debug(f"Signature from {signing_key_log} for {key_id} verified")
                        await self.crypto_store.put_signature(
                            target=CrossSigner(user_id, actual_key),
                            signer=CrossSigner(signer_user_id, signing_key),
                            signature=signature,
                        )
                    else:
                        self.log.warning(f"Invalid signature from {signing_key_log} for {key_id}")

    async def get_or_fetch_device(
        self, user_id: UserID, device_id: DeviceID
    ) -> DeviceIdentity | None:
        device = await self.crypto_store.get_device(user_id, device_id)
        if device is not None:
            return device
        devices = await self._fetch_keys([user_id], include_untracked=True)
        try:
            return devices[user_id][device_id]
        except KeyError:
            return None

    async def get_or_fetch_device_by_key(
        self, user_id: UserID, identity_key: IdentityKey
    ) -> DeviceIdentity | None:
        device = await self.crypto_store.find_device_by_key(user_id, identity_key)
        if device is not None:
            return device
        devices = await self._fetch_keys([user_id], include_untracked=True)
        for device in devices.get(user_id, {}).values():
            if device.identity_key == identity_key:
                return device
        return None

    async def on_devices_changed(self, user_id: UserID) -> None:
        if self.disable_device_change_key_rotation:
            return
        shared_rooms = await self.state_store.find_shared_rooms(user_id)
        self.log.debug(
            f"Devices of {user_id} changed, invalidating group session in {shared_rooms}"
        )
        await self.crypto_store.remove_outbound_group_sessions(shared_rooms)

    @staticmethod
    async def _validate_device(
        user_id: UserID,
        device_id: DeviceID,
        device_keys: DeviceKeys,
        existing: DeviceIdentity | None = None,
    ) -> DeviceIdentity:
        if user_id != device_keys.user_id:
            raise DeviceValidationError(
                f"mismatching user ID (expected {user_id}, got {device_keys.user_id})"
            )
        elif device_id != device_keys.device_id:
            raise DeviceValidationError(
                f"mismatching device ID (expected {device_id}, got {device_keys.device_id})"
            )

        signing_key = device_keys.ed25519
        if not signing_key:
            raise DeviceValidationError("didn't find ed25519 signing key")
        identity_key = device_keys.curve25519
        if not identity_key:
            raise DeviceValidationError("didn't find curve25519 identity key")

        if existing and existing.signing_key != signing_key:
            raise DeviceValidationError(
                f"received update for device with different signing key "
                f"(expected {existing.signing_key}, got {signing_key})"
            )

        if not verify_signature_json(device_keys.serialize(), user_id, device_id, signing_key):
            raise DeviceValidationError("invalid signature on device keys")

        name = device_keys.unsigned.device_display_name or device_id

        return DeviceIdentity(
            user_id=user_id,
            device_id=device_id,
            identity_key=identity_key,
            signing_key=signing_key,
            trust=TrustState.UNVERIFIED,
            name=name,
            deleted=False,
        )

    async def resolve_trust(self, device: DeviceIdentity) -> TrustState:
        try:
            return await self._try_resolve_trust(device)
        except Exception:
            self.log.exception(f"Failed to resolve trust of {device.user_id}/{device.device_id}")
            return TrustState.UNVERIFIED

    async def _try_resolve_trust(self, device: DeviceIdentity) -> TrustState:
        if device.trust in (TrustState.VERIFIED, TrustState.BLACKLISTED):
            return device.trust
        their_keys = await self.crypto_store.get_cross_signing_keys(device.user_id)
        if len(their_keys) == 0 and device.user_id not in self._cs_fetch_attempted:
            self.log.debug(f"Didn't find any cross-signing keys for {device.user_id}, fetching...")
            async with self._fetch_keys_lock:
                if device.user_id not in self._cs_fetch_attempted:
                    self._cs_fetch_attempted.add(device.user_id)
                    await self._fetch_keys([device.user_id])
            their_keys = await self.crypto_store.get_cross_signing_keys(device.user_id)
        try:
            msk = their_keys[CrossSigningUsage.MASTER]
            ssk = their_keys[CrossSigningUsage.SELF]
        except KeyError as e:
            self.log.error(f"Didn't find cross-signing key {e.args[0]} of {device.user_id}")
            return TrustState.UNVERIFIED
        ssk_signed = await self.crypto_store.is_key_signed_by(
            target=CrossSigner(device.user_id, ssk.key),
            signer=CrossSigner(device.user_id, msk.key),
        )
        if not ssk_signed:
            self.log.warning(
                f"Self-signing key of {device.user_id} is not signed by their master key"
            )
            return TrustState.UNVERIFIED
        device_signed = await self.crypto_store.is_key_signed_by(
            target=CrossSigner(device.user_id, device.signing_key),
            signer=CrossSigner(device.user_id, ssk.key),
        )
        if device_signed:
            if await self.is_user_trusted(device.user_id):
                return TrustState.CROSS_SIGNED_TRUSTED
            elif msk.key == msk.first:
                return TrustState.CROSS_SIGNED_TOFU
            return TrustState.CROSS_SIGNED_UNTRUSTED
        return TrustState.UNVERIFIED

    async def is_user_trusted(self, user_id: UserID) -> bool:
        # TODO implement once own cross-signing key stuff is ready
        return False



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from mautrix.api import Method, Path
from mautrix.types import (
    PushAction,
    PushCondition,
    PushRule,
    PushRuleID,
    PushRuleKind,
    PushRuleScope,
)

from ..base import BaseClientAPI


class PushRuleMethods(BaseClientAPI):
    """
    Methods in section 13.13 Push Notifications of the spec. These methods are used for modifying
    what triggers push notifications.

    See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#id89>`__"""

    async def get_push_rule(
        self, scope: PushRuleScope, kind: PushRuleKind, rule_id: PushRuleID
    ) -> PushRule:
        """
        Retrieve a single specified push rule.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#get-matrix-client-r0-pushrules-scope-kind-ruleid>`__

        Args:
            scope: The scope of the push rule.
            kind: The kind of rule.
            rule_id: The identifier of the rule.

        Returns:
            The push rule information.
        """
        resp = await self.api.request(Method.GET, Path.v3.pushrules[scope][kind][rule_id])
        return PushRule.deserialize(resp)

    async def set_push_rule(
        self,
        scope: PushRuleScope,
        kind: PushRuleKind,
        rule_id: PushRuleID,
        actions: list[PushAction],
        pattern: str | None = None,
        before: PushRuleID | None = None,
        after: PushRuleID | None = None,
        conditions: list[PushCondition] = None,
    ) -> None:
        """
        Create or modify a push rule.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#put-matrix-client-r0-pushrules-scope-kind-ruleid>`__

        Args:
            scope: The scope of the push rule.
            kind: The kind of rule.
            rule_id: The identifier for the rule.
            before:
            after:
            actions: The actions to perform when the conditions for the rule are met.
            pattern: The glob-style pattern to match against for ``content`` rules.
            conditions: The conditions for the rule for ``underride`` and ``override`` rules.
        """
        query = {}
        if after:
            query["after"] = after
        if before:
            query["before"] = before
        content = {"actions": [act.serialize() for act in actions]}
        if conditions:
            content["conditions"] = [cond.serialize() for cond in conditions]
        if pattern:
            content["pattern"] = pattern
        await self.api.request(
            Method.PUT,
            Path.v3.pushrules[scope][kind][rule_id],
            query_params=query,
            content=content,
        )

    async def remove_push_rule(
        self, scope: PushRuleScope, kind: PushRuleKind, rule_id: PushRuleID
    ) -> None:
        """
        Remove a push rule.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#delete-matrix-client-r0-pushrules-scope-kind-ruleid>`__

        Args:
            scope: The scope of the push rule.
            kind: The kind of rule.
            rule_id: The identifier of the rule.
        """
        await self.api.request(Method.DELETE, Path.v3.pushrules[scope][kind][rule_id])



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import json

import olm

from mautrix.errors import (
    DecryptedPayloadError,
    DecryptionError,
    DuplicateMessageIndex,
    MismatchingRoomError,
    SessionNotFound,
    VerificationError,
)
from mautrix.types import (
    EncryptedEvent,
    EncryptedMegolmEventContent,
    EncryptionAlgorithm,
    Event,
    SessionID,
    TrustState,
)

from .device_lists import DeviceListMachine
from .sessions import InboundGroupSession


class MegolmDecryptionMachine(DeviceListMachine):
    async def decrypt_megolm_event(self, evt: EncryptedEvent) -> Event:
        """
        Decrypt an event that was encrypted using Megolm.

        Args:
            evt: The whole encrypted event.

        Returns:
            The decrypted event, including some unencrypted metadata from the input event.

        Raises:
            DecryptionError: If decryption failed.
        """
        if not isinstance(evt.content, EncryptedMegolmEventContent):
            raise DecryptionError("Unsupported event content class")
        elif evt.content.algorithm != EncryptionAlgorithm.MEGOLM_V1:
            raise DecryptionError("Unsupported event encryption algorithm")
        async with self._megolm_decrypt_lock:
            session = await self.crypto_store.get_group_session(
                evt.room_id, evt.content.session_id
            )
            if session is None:
                # TODO check if olm session is wedged
                raise SessionNotFound(evt.content.session_id, evt.content.sender_key)
            try:
                plaintext, index = session.decrypt(evt.content.ciphertext)
            except olm.OlmGroupSessionError as e:
                raise DecryptionError("Failed to decrypt megolm event") from e
            if not await self.crypto_store.validate_message_index(
                session.sender_key, SessionID(session.id), evt.event_id, index, evt.timestamp
            ):
                raise DuplicateMessageIndex()
            await self._ratchet_session(session, index)

        forwarded_keys = False
        if (
            evt.content.device_id == self.client.device_id
            and session.signing_key == self.account.signing_key
            and session.sender_key == self.account.identity_key
            and not session.forwarding_chain
        ):
            trust_level = TrustState.VERIFIED
        else:
            device = await self.get_or_fetch_device_by_key(evt.sender, session.sender_key)
            if not session.forwarding_chain or (
                len(session.forwarding_chain) == 1
                and session.forwarding_chain[0] == session.sender_key
            ):
                if not device:
                    self.log.debug(
                        f"Couldn't resolve trust level of session {session.id}: "
                        f"sent by unknown device {evt.sender}/{session.sender_key}"
                    )
                    trust_level = TrustState.UNKNOWN_DEVICE
                elif (
                    device.signing_key != session.signing_key
                    or device.identity_key != session.sender_key
                ):
                    raise VerificationError()
                else:
                    trust_level = await self.resolve_trust(device)
            else:
                forwarded_keys = True
                last_chain_item = session.forwarding_chain[-1]
                received_from = await self.crypto_store.find_device_by_key(
                    evt.sender, last_chain_item
                )
                if received_from:
                    trust_level = await self.resolve_trust(received_from)
                else:
                    self.log.debug(
                        f"Couldn't resolve trust level of session {session.id}: "
                        f"forwarding chain ends with unknown device {last_chain_item}"
                    )
                    trust_level = TrustState.FORWARDED

        try:
            data = json.loads(plaintext)
            room_id = data["room_id"]
            event_type = data["type"]
            content = data["content"]
        except json.JSONDecodeError as e:
            raise DecryptedPayloadError("Failed to parse megolm payload") from e
        except KeyError as e:
            raise DecryptedPayloadError("Megolm payload is missing fields") from e

        if room_id != evt.room_id:
            raise MismatchingRoomError()

        if evt.content.relates_to and "m.relates_to" not in content:
            content["m.relates_to"] = evt.content.relates_to.serialize()
        result = Event.deserialize(
            {
                "room_id": evt.room_id,
                "event_id": evt.event_id,
                "sender": evt.sender,
                "origin_server_ts": evt.timestamp,
                "type": event_type,
                "content": content,
            }
        )
        result.unsigned = evt.unsigned
        result.type = result.type.with_class(evt.type.t_class)
        result["mautrix"] = {
            "trust_state": trust_level,
            "forwarded_keys": forwarded_keys,
            "was_encrypted": True,
        }
        return result

    async def _ratchet_session(self, sess: InboundGroupSession, index: int) -> None:
        expected_message_index = sess.ratchet_safety.next_index
        did_modify = True
        if index > expected_message_index:
            sess.ratchet_safety.missed_indices += list(range(expected_message_index, index))
            sess.ratchet_safety.next_index = index + 1
        elif index == expected_message_index:
            sess.ratchet_safety.next_index = index + 1
        else:
            try:
                sess.ratchet_safety.missed_indices.remove(index)
            except ValueError:
                did_modify = False
        # Use presence of received_at as a sign that this is a recent megolm session,
        # and therefore it's safe to drop missed indices entirely.
        if (
            sess.received_at
            and sess.ratchet_safety.missed_indices
            and sess.ratchet_safety.missed_indices[0] < expected_message_index - 10
        ):
            i = 0
            for i, lost_index in enumerate(sess.ratchet_safety.missed_indices):
                if lost_index < expected_message_index - 10:
                    sess.ratchet_safety.lost_indices.append(lost_index)
                else:
                    break
            sess.ratchet_safety.missed_indices = sess.ratchet_safety.missed_indices[i + 1 :]
        ratchet_target_index = sess.ratchet_safety.next_index
        if len(sess.ratchet_safety.missed_indices) > 0:
            ratchet_target_index = min(sess.ratchet_safety.missed_indices)
        self.log.debug(
            f"Ratchet safety info for {sess.id}: {sess.ratchet_safety}, {ratchet_target_index=}"
        )
        sess_id = SessionID(sess.id)
        if (
            sess.max_messages
            and ratchet_target_index >= sess.max_messages
            and not sess.ratchet_safety.missed_indices
            and self.delete_fully_used_keys_on_decrypt
        ):
            self.log.info(f"Deleting fully used session {sess.id}")
            await self.crypto_store.redact_group_session(
                sess.room_id, sess_id, reason="maximum messages reached"
            )
            return
        elif sess.first_known_index < ratchet_target_index and self.ratchet_keys_on_decrypt:
            self.log.info(f"Ratcheting session {sess.id} to {ratchet_target_index}")
            sess = sess.ratchet_to(ratchet_target_index)
        elif not did_modify:
            return
        await self.crypto_store.put_group_session(sess.room_id, sess.sender_key, sess_id, sess)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any
from abc import ABC, abstractmethod
from enum import Enum
import asyncio
import sys

from aiohttp import web

from mautrix import __version__ as __mautrix_version__
from mautrix.api import HTTPAPI
from mautrix.appservice import AppService, ASStateStore
from mautrix.client.state_store.asyncpg import PgStateStore as PgClientStateStore
from mautrix.errors import MExclusive, MUnknownToken
from mautrix.types import RoomID, UserID
from mautrix.util.async_db import Database, DatabaseException, UpgradeTable
from mautrix.util.bridge_state import BridgeState, BridgeStateEvent, GlobalBridgeState
from mautrix.util.program import Program

from .. import bridge as br
from .state_store.asyncpg import PgBridgeStateStore

try:
    import uvloop
except ImportError:
    uvloop = None


class HomeserverSoftware(Enum):
    STANDARD = "standard"
    ASMUX = "asmux"
    HUNGRY = "hungry"

    @property
    def is_hungry(self) -> bool:
        return self == self.HUNGRY

    @property
    def is_asmux(self) -> bool:
        return self == self.ASMUX


class Bridge(Program, ABC):
    db: Database
    az: AppService
    state_store_class: type[ASStateStore] = PgBridgeStateStore
    state_store: ASStateStore
    upgrade_table: UpgradeTable
    config_class: type[br.BaseBridgeConfig]
    config: br.BaseBridgeConfig
    matrix_class: type[br.BaseMatrixHandler]
    matrix: br.BaseMatrixHandler
    repo_url: str
    markdown_version: str
    manhole: br.commands.manhole.ManholeState | None
    homeserver_software: HomeserverSoftware
    beeper_network_name: str | None = None
    beeper_service_name: str | None = None

    def __init__(
        self,
        module: str = None,
        name: str = None,
        description: str = None,
        command: str = None,
        version: str = None,
        config_class: type[br.BaseBridgeConfig] = None,
        matrix_class: type[br.BaseMatrixHandler] = None,
        state_store_class: type[ASStateStore] = None,
    ) -> None:
        super().__init__(module, name, description, command, version, config_class)
        if matrix_class:
            self.matrix_class = matrix_class
        if state_store_class:
            self.state_store_class = state_store_class
        self.manhole = None

    def prepare_arg_parser(self) -> None:
        super().prepare_arg_parser()
        self.parser.add_argument(
            "-g",
            "--generate-registration",
            action="store_true",
            help="generate registration and quit",
        )
        self.parser.add_argument(
            "-r",
            "--registration",
            type=str,
            default="registration.yaml",
            metavar="<path>",
            help=(
                "the path to save the generated registration to "
                "(not needed for running the bridge)"
            ),
        )
        self.parser.add_argument(
            "--ignore-unsupported-database",
            action="store_true",
            help="Run even if the database schema is too new",
        )
        self.parser.add_argument(
            "--ignore-foreign-tables",
            action="store_true",
            help="Run even if the database contains tables from other programs (like Synapse)",
        )

    def preinit(self) -> None:
        super().preinit()
        if self.args.generate_registration:
            self.generate_registration()
            sys.exit(0)

    def prepare(self) -> None:
        if self.config.env:
            self.log.debug(
                "Loaded config overrides from environment: %s", list(self.config.env.keys())
            )
        super().prepare()
        try:
            self.homeserver_software = HomeserverSoftware(self.config["homeserver.software"])
        except Exception:
            self.log.fatal("Invalid value for homeserver.software in config")
            sys.exit(11)
        self.prepare_db()
        self.prepare_appservice()
        self.prepare_bridge()

    def prepare_config(self) -> None:
        self.config = self.config_class(
            self.args.config,
            self.args.registration,
            self.args.base_config,
            env_prefix=self.module.upper(),
        )
        if self.args.generate_registration:
            self.config._check_tokens = False
        self.load_and_update_config()

    def generate_registration(self) -> None:
        self.config.generate_registration()
        self.config.save()
        print(f"Registration generated and saved to {self.config.registration_path}")

    def make_state_store(self) -> None:
        if self.state_store_class is None:
            raise RuntimeError("state_store_class is not set")
        elif issubclass(self.state_store_class, PgBridgeStateStore):
            self.state_store = self.state_store_class(
                self.db, self.get_puppet, self.get_double_puppet
            )
        else:
            self.state_store = self.state_store_class()

    def prepare_appservice(self) -> None:
        self.make_state_store()
        mb = 1024**2
        default_http_retry_count = self.config.get("homeserver.http_retry_count", None)
        if self.name not in HTTPAPI.default_ua:
            HTTPAPI.default_ua = f"{self.name}/{self.version} {HTTPAPI.default_ua}"
        self.az = AppService(
            server=self.config["homeserver.address"],
            domain=self.config["homeserver.domain"],
            verify_ssl=self.config["homeserver.verify_ssl"],
            connection_limit=self.config["homeserver.connection_limit"],
            id=self.config["appservice.id"],
            as_token=self.config["appservice.as_token"],
            hs_token=self.config["appservice.hs_token"],
            tls_cert=self.config.get("appservice.tls_cert", None),
            tls_key=self.config.get("appservice.tls_key", None),
            bot_localpart=self.config["appservice.bot_username"],
            ephemeral_events=self.config["appservice.ephemeral_events"],
            encryption_events=self.config["bridge.encryption.appservice"],
            default_ua=HTTPAPI.default_ua,
            default_http_retry_count=default_http_retry_count,
            log="mau.as",
            loop=self.loop,
            state_store=self.state_store,
            bridge_name=self.name,
            aiohttp_params={"client_max_size": self.config["appservice.max_body_size"] * mb},
        )
        self.az.app.router.add_post("/_matrix/app/com.beeper.bridge_state", self.get_bridge_state)

    def prepare_db(self) -> None:
        if not hasattr(self, "upgrade_table") or not self.upgrade_table:
            raise RuntimeError("upgrade_table is not set")
        self.db = Database.create(
            self.config["appservice.database"],
            upgrade_table=self.upgrade_table,
            db_args=self.config["appservice.database_opts"],
            owner_name=self.name,
            ignore_foreign_tables=self.args.ignore_foreign_tables,
        )

    def prepare_bridge(self) -> None:
        self.matrix = self.matrix_class(bridge=self)

    def _log_db_error(self, e: Exception) -> None:
        self.log.critical("Failed to initialize database", exc_info=e)
        if isinstance(e, DatabaseException) and e.explanation:
            self.log.info(e.explanation)
        sys.exit(25)

    async def start_db(self) -> None:
        if hasattr(self, "db") and isinstance(self.db, Database):
            self.log.debug("Starting database...")
            ignore_unsupported = self.args.ignore_unsupported_database
            self.db.upgrade_table.allow_unsupported = ignore_unsupported
            try:
                await self.db.start()
                if isinstance(self.state_store, PgClientStateStore):
                    self.state_store.upgrade_table.allow_unsupported = ignore_unsupported
                    await self.state_store.upgrade_table.upgrade(self.db)
                if self.matrix.e2ee:
                    self.matrix.e2ee.crypto_db.allow_unsupported = ignore_unsupported
                    self.matrix.e2ee.crypto_db.override_pool(self.db)
            except Exception as e:
                self._log_db_error(e)

    async def stop_db(self) -> None:
        if hasattr(self, "db") and isinstance(self.db, Database):
            await self.db.stop()

    async def start(self) -> None:
        await self.start_db()

        self.log.debug("Starting appservice...")
        await self.az.start(self.config["appservice.hostname"], self.config["appservice.port"])
        try:
            await self.matrix.wait_for_connection()
        except MUnknownToken:
            self.log.critical(
                "The as_token was not accepted. Is the registration file installed "
                "in your homeserver correctly?"
            )
            sys.exit(16)
        except MExclusive:
            self.log.critical(
                "The as_token was accepted, but the /register request was not. "
                "Are the homeserver domain and username template in the config "
                "correct, and do they match the values in the registration?"
            )
            sys.exit(16)
        except Exception:
            self.log.critical("Failed to check connection to homeserver", exc_info=True)
            sys.exit(16)

        await self.matrix.init_encryption()
        self.add_startup_actions(self.matrix.init_as_bot())
        await super().start()
        self.az.ready = True

        status_endpoint = self.config["homeserver.status_endpoint"]
        if status_endpoint and await self.count_logged_in_users() == 0:
            state = BridgeState(state_event=BridgeStateEvent.UNCONFIGURED).fill()
            while not await state.send(status_endpoint, self.az.as_token, self.log):
                await asyncio.sleep(5)

    async def system_exit(self) -> None:
        if hasattr(self, "db") and isinstance(self.db, Database):
            self.log.debug("Stopping database due to SystemExit")
            await self.db.stop()
            self.log.debug("Database stopped")
        elif getattr(self, "db", None):
            self.log.trace("Database not started at SystemExit")

    async def stop(self) -> None:
        if self.manhole:
            self.manhole.close()
            self.manhole = None
        await self.az.stop()
        await super().stop()
        if self.matrix.e2ee:
            await self.matrix.e2ee.stop()
        await self.stop_db()

    async def get_bridge_state(self, req: web.Request) -> web.Response:
        if not self.az._check_token(req):
            return web.json_response({"error": "Invalid auth token"}, status=401)
        try:
            user = await self.get_user(UserID(req.url.query["user_id"]), create=False)
        except KeyError:
            user = None
        if user is None:
            return web.json_response({"error": "User not found"}, status=404)
        try:
            states = await user.get_bridge_states()
        except NotImplementedError:
            return web.json_response({"error": "Bridge status not implemented"}, status=501)
        for state in states:
            await user.fill_bridge_state(state)
        global_state = BridgeState(state_event=BridgeStateEvent.RUNNING).fill()
        evt = GlobalBridgeState(
            remote_states={state.remote_id: state for state in states}, bridge_state=global_state
        )
        return web.json_response(evt.serialize())

    @abstractmethod
    async def get_user(self, user_id: UserID, create: bool = True) -> br.BaseUser | None:
        pass

    @abstractmethod
    async def get_portal(self, room_id: RoomID) -> br.BasePortal | None:
        pass

    @abstractmethod
    async def get_puppet(self, user_id: UserID, create: bool = False) -> br.BasePuppet | None:
        pass

    @abstractmethod
    async def get_double_puppet(self, user_id: UserID) -> br.BasePuppet | None:
        pass

    @abstractmethod
    def is_bridge_ghost(self, user_id: UserID) -> bool:
        pass

    @abstractmethod
    async def count_logged_in_users(self) -> int:
        return 0

    async def manhole_global_namespace(self, user_id: UserID) -> dict[str, Any]:
        own_user = await self.get_user(user_id, create=False)
        try:
            own_puppet = await own_user.get_puppet()
        except NotImplementedError:
            own_puppet = None
        return {
            "bridge": self,
            "manhole": self.manhole,
            "own_user": own_user,
            "own_puppet": own_puppet,
        }

    @property
    def manhole_banner_python_version(self) -> str:
        return f"Python {sys.version} on {sys.platform}"

    @property
    def manhole_banner_program_version(self) -> str:
        return f"{self.name} {self.version} with mautrix-python {__mautrix_version__}"

    def manhole_banner(self, user_id: UserID) -> str:
        return (
            f"{self.manhole_banner_python_version}\n"
            f"{self.manhole_banner_program_version}\n\n"
            f"Manhole opened by {user_id}\n"
        )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


class MatrixError(Exception):
    """A generic Matrix error. Specific errors will subclass this."""

    pass


class MatrixConnectionError(MatrixError):
    pass


class MatrixResponseError(MatrixError):
    """The response from the homeserver did not fulfill expectations."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class IntentError(MatrixError):
    """An intent execution failure, most likely caused by a `MatrixRequestError`."""

    pass



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, cast


class _NoopPrometheusEntity:
    """NoopPrometheusEntity is a class that can be used as a no-op placeholder for prometheus
    metrics objects when prometheus_client isn't installed."""

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        if not kwargs and len(args) == 1 and callable(args[0]):
            return args[0]
        return self

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __getattr__(self, item):
        return self


try:
    from prometheus_client import Counter, Enum, Gauge, Histogram, Info, Summary

    is_installed = True
except ImportError:
    Counter = Gauge = Summary = Histogram = Info = Enum = cast(Any, _NoopPrometheusEntity)

    is_installed = False


def async_time(metric: Gauge | Summary | Histogram):
    """
    Measure the time that each execution of the decorated async function takes.

    This is equivalent to the ``time`` method-decorator in the metrics, but
    supports async functions.

    Args:
        metric: The metric instance to store the measures in.
    """
    if not hasattr(metric, "time") or not callable(metric.time):
        raise ValueError("async_time only supports metrics that support timing")

    def decorator(fn):
        async def wrapper(*args, **kwargs):
            with metric.time():
                return await fn(*args, **kwargs)

        return wrapper if is_installed else fn

    return decorator


__all__ = [
    "Counter",
    "Gauge",
    "Summary",
    "Histogram",
    "Info",
    "Enum",
    "async_time",
    "is_installed",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Generic, Iterable, Sequence, Type, TypeVar
from abc import ABC, abstractmethod
from itertools import chain

from attr import dataclass
import attr

from .formatted_string import EntityType, FormattedString


class AbstractEntity(ABC):
    def __init__(
        self, type: EntityType, offset: int, length: int, extra_info: dict[str, Any]
    ) -> None:
        pass

    @abstractmethod
    def copy(self) -> AbstractEntity:
        pass

    @abstractmethod
    def adjust_offset(self, offset: int, max_length: int = -1) -> AbstractEntity | None:
        pass


class SemiAbstractEntity(AbstractEntity, ABC):
    offset: int
    length: int

    def adjust_offset(self, offset: int, max_length: int = -1) -> SemiAbstractEntity | None:
        entity = self.copy()
        entity.offset += offset
        if entity.offset < 0:
            entity.length += entity.offset
            if entity.length < 0:
                return None
            entity.offset = 0
        elif entity.offset > max_length > -1:
            return None
        elif entity.offset + entity.length > max_length > -1:
            entity.length = max_length - entity.offset
        return entity


@dataclass
class SimpleEntity(SemiAbstractEntity):
    type: EntityType
    offset: int
    length: int
    extra_info: dict[str, Any] = attr.ib(factory=dict)

    def copy(self) -> SimpleEntity:
        return attr.evolve(self)


TEntity = TypeVar("TEntity", bound=AbstractEntity)
TEntityType = TypeVar("TEntityType")


class EntityString(Generic[TEntity, TEntityType], FormattedString):
    text: str
    _entities: list[TEntity]
    entity_class: Type[AbstractEntity] = SimpleEntity

    def __init__(self, text: str = "", entities: list[TEntity] = None) -> None:
        self.text = text
        self._entities = entities or []

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text='{self.text}', entities={self.entities})"

    def __str__(self) -> str:
        return self.text

    @property
    def entities(self) -> list[TEntity]:
        return self._entities

    @entities.setter
    def entities(self, val: Iterable[TEntity]) -> None:
        self._entities = [entity for entity in val if entity is not None]

    def _offset_entities(self, offset: int) -> EntityString:
        self.entities = (entity.adjust_offset(offset, len(self.text)) for entity in self.entities)
        return self

    def append(self, *args: str | FormattedString) -> EntityString:
        for msg in args:
            if isinstance(msg, EntityString):
                self.entities += (entity.adjust_offset(len(self.text)) for entity in msg.entities)
                self.text += msg.text
            else:
                self.text += str(msg)
        return self

    def prepend(self, *args: str | FormattedString) -> EntityString:
        for msg in args:
            if isinstance(msg, EntityString):
                self.text = msg.text + self.text
                self.entities = chain(
                    msg.entities, (entity.adjust_offset(len(msg.text)) for entity in self.entities)
                )
            else:
                text = str(msg)
                self.text = text + self.text
                self.entities = (entity.adjust_offset(len(text)) for entity in self.entities)
        return self

    def format(
        self, entity_type: TEntityType, offset: int = None, length: int = None, **kwargs
    ) -> EntityString:
        self.entities.append(
            self.entity_class(
                type=entity_type,
                offset=offset or 0,
                length=length or len(self.text),
                extra_info=kwargs,
            )
        )
        return self

    def trim(self) -> EntityString:
        orig_len = len(self.text)
        self.text = self.text.lstrip()
        diff = orig_len - len(self.text)
        self.text = self.text.rstrip()
        self._offset_entities(-diff)
        return self

    def split(self, separator, max_items: int = -1) -> list[EntityString]:
        text_parts = self.text.split(separator, max_items - 1)
        output: list[EntityString] = []

        offset = 0
        for part in text_parts:
            msg = type(self)(part)
            msg.entities = (entity.adjust_offset(-offset, len(part)) for entity in self.entities)
            output.append(msg)

            offset += len(part)
            offset += len(separator)

        return output

    @classmethod
    def join(cls, items: Sequence[str | EntityString], separator: str = " ") -> EntityString:
        main = cls()
        for msg in items:
            if not isinstance(msg, EntityString):
                msg = cls(text=str(msg))
            main.entities += [entity.adjust_offset(len(main.text)) for entity in msg.entities]
            main.text += msg.text + separator
        if len(separator) > 0:
            main.text = main.text[: -len(separator)]
        return main



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Optional, Union

from attr import dataclass
import attr

from ..primitive import JSON, ContentURI, EventID, RoomAlias, RoomID, UserID
from ..util import Obj, SerializableAttrs, SerializableEnum, deserializer, field
from .base import BaseRoomEvent, BaseUnsigned
from .encrypted import EncryptionAlgorithm
from .type import EventType, RoomType


@dataclass
class NotificationPowerLevels(SerializableAttrs):
    room: int = 50


@dataclass
class PowerLevelStateEventContent(SerializableAttrs):
    """The content of a power level event."""

    users: Dict[UserID, int] = attr.ib(default=attr.Factory(dict), metadata={"omitempty": False})
    users_default: int = 0

    events: Dict[EventType, int] = attr.ib(
        default=attr.Factory(dict), metadata={"omitempty": False}
    )
    events_default: int = 0

    notifications: NotificationPowerLevels = attr.ib(factory=lambda: NotificationPowerLevels())

    state_default: int = 50

    invite: int = 50
    kick: int = 50
    ban: int = 50
    redact: int = 50

    def get_user_level(self, user_id: UserID) -> int:
        return int(self.users.get(user_id, self.users_default))

    def set_user_level(self, user_id: UserID, level: int) -> None:
        if level == self.users_default:
            del self.users[user_id]
        else:
            self.users[user_id] = level

    def ensure_user_level(self, user_id: UserID, level: int) -> bool:
        if self.get_user_level(user_id) != level:
            self.set_user_level(user_id, level)
            return True
        return False

    def get_event_level(self, event_type: EventType) -> int:
        return int(
            self.events.get(
                event_type, (self.state_default if event_type.is_state else self.events_default)
            )
        )

    def set_event_level(self, event_type: EventType, level: int) -> None:
        if level == self.state_default if event_type.is_state else self.events_default:
            del self.events[event_type]
        else:
            self.events[event_type] = level

    def ensure_event_level(self, event_type: EventType, level: int) -> bool:
        if self.get_event_level(event_type) != level:
            self.set_event_level(event_type, level)
            return True
        return False


class Membership(SerializableEnum):
    """
    The membership state of a user in a room as specified in section `8.4 Room membership`_ of the
    spec.

    .. _8.4 Room membership: https://spec.matrix.org/v1.2/client-server-api/#room-membership
    """

    JOIN = "join"
    LEAVE = "leave"
    INVITE = "invite"
    BAN = "ban"
    KNOCK = "knock"


@dataclass
class MemberStateEventContent(SerializableAttrs):
    """The content of a membership event. `Spec link`_

    .. _Spec link: https://spec.matrix.org/v1.2/client-server-api/#mroommember"""

    membership: Membership = Membership.LEAVE
    avatar_url: ContentURI = None
    displayname: str = None
    is_direct: bool = False
    reason: str = None
    third_party_invite: JSON = None


@dataclass
class CanonicalAliasStateEventContent(SerializableAttrs):
    """
    The content of a ``m.room.canonical_alias`` event (:attr:`EventType.ROOM_CANONICAL_ALIAS`).

    This event is used to inform the room about which alias should be considered the canonical one,
    and which other aliases point to the room. This could be for display purposes or as suggestion
    to users which alias to use to advertise and access the room.

    See also: `m.room.canonical_alias in the spec`_

    .. _m.room.canonical_alias in the spec: https://spec.matrix.org/v1.2/client-server-api/#mroomcanonical_alias
    """

    canonical_alias: RoomAlias = attr.ib(default=None, metadata={"json": "alias"})
    alt_aliases: List[RoomAlias] = attr.ib(factory=lambda: [])


@dataclass
class RoomNameStateEventContent(SerializableAttrs):
    name: str = None


@dataclass
class RoomTopicStateEventContent(SerializableAttrs):
    topic: str = None


@dataclass
class RoomAvatarStateEventContent(SerializableAttrs):
    url: Optional[ContentURI] = None


class JoinRule(SerializableEnum):
    PUBLIC = "public"
    KNOCK = "knock"
    RESTRICTED = "restricted"
    INVITE = "invite"
    PRIVATE = "private"
    KNOCK_RESTRICTED = "knock_restricted"


class JoinRestrictionType(SerializableEnum):
    ROOM_MEMBERSHIP = "m.room_membership"


@dataclass
class JoinRestriction(SerializableAttrs):
    type: JoinRestrictionType
    room_id: Optional[RoomID] = None


@dataclass
class JoinRulesStateEventContent(SerializableAttrs):
    join_rule: JoinRule
    allow: Optional[List[JoinRestriction]] = None


@dataclass
class RoomPinnedEventsStateEventContent(SerializableAttrs):
    pinned: List[EventID] = None


@dataclass
class RoomTombstoneStateEventContent(SerializableAttrs):
    body: str = None
    replacement_room: RoomID = None


@dataclass
class RoomEncryptionStateEventContent(SerializableAttrs):
    algorithm: EncryptionAlgorithm = None
    rotation_period_ms: int = 604800000
    rotation_period_msgs: int = 100


@dataclass
class RoomPredecessor(SerializableAttrs):
    room_id: RoomID = None
    event_id: EventID = None


@dataclass
class RoomCreateStateEventContent(SerializableAttrs):
    room_version: str = "1"
    federate: bool = field(json="m.federate", omit_default=True, default=True)
    predecessor: Optional[RoomPredecessor] = None
    type: Optional[RoomType] = None


@dataclass
class SpaceChildStateEventContent(SerializableAttrs):
    via: List[str] = None
    order: str = ""
    suggested: bool = False


@dataclass
class SpaceParentStateEventContent(SerializableAttrs):
    via: List[str] = None
    canonical: bool = False


StateEventContent = Union[
    PowerLevelStateEventContent,
    MemberStateEventContent,
    CanonicalAliasStateEventContent,
    RoomNameStateEventContent,
    RoomAvatarStateEventContent,
    RoomTopicStateEventContent,
    RoomPinnedEventsStateEventContent,
    RoomTombstoneStateEventContent,
    RoomEncryptionStateEventContent,
    RoomCreateStateEventContent,
    SpaceChildStateEventContent,
    SpaceParentStateEventContent,
    JoinRulesStateEventContent,
    Obj,
]


@dataclass
class StrippedStateUnsigned(BaseUnsigned, SerializableAttrs):
    """Unsigned information sent with state events."""

    prev_content: StateEventContent = None
    prev_sender: UserID = None
    replaces_state: EventID = None


@dataclass
class StrippedStateEvent(SerializableAttrs):
    """Stripped state events included with some invite events."""

    content: StateEventContent = None
    room_id: RoomID = None
    sender: UserID = None
    type: EventType = None
    state_key: str = None

    unsigned: Optional[StrippedStateUnsigned] = None

    @property
    def prev_content(self) -> StateEventContent:
        if self.unsigned and self.unsigned.prev_content:
            return self.unsigned.prev_content
        return state_event_content_map.get(self.type, Obj)()

    @classmethod
    def deserialize(cls, data: JSON) -> "StrippedStateEvent":
        try:
            event_type = EventType.find(data.get("type", None))
            data.get("content", {})["__mautrix_event_type"] = event_type
            (data.get("unsigned") or {}).get("prev_content", {})[
                "__mautrix_event_type"
            ] = event_type
        except ValueError:
            pass
        return super().deserialize(data)


@dataclass
class StateUnsigned(StrippedStateUnsigned, SerializableAttrs):
    invite_room_state: Optional[List[StrippedStateEvent]] = None


state_event_content_map = {
    EventType.ROOM_CREATE: RoomCreateStateEventContent,
    EventType.ROOM_POWER_LEVELS: PowerLevelStateEventContent,
    EventType.ROOM_MEMBER: MemberStateEventContent,
    EventType.ROOM_PINNED_EVENTS: RoomPinnedEventsStateEventContent,
    EventType.ROOM_CANONICAL_ALIAS: CanonicalAliasStateEventContent,
    EventType.ROOM_NAME: RoomNameStateEventContent,
    EventType.ROOM_AVATAR: RoomAvatarStateEventContent,
    EventType.ROOM_TOPIC: RoomTopicStateEventContent,
    EventType.ROOM_JOIN_RULES: JoinRulesStateEventContent,
    EventType.ROOM_TOMBSTONE: RoomTombstoneStateEventContent,
    EventType.ROOM_ENCRYPTION: RoomEncryptionStateEventContent,
    EventType.SPACE_CHILD: SpaceChildStateEventContent,
    EventType.SPACE_PARENT: SpaceParentStateEventContent,
}


@dataclass
class StateEvent(BaseRoomEvent, SerializableAttrs):
    """A room state event."""

    state_key: str
    content: StateEventContent
    unsigned: Optional[StateUnsigned] = field(factory=lambda: StateUnsigned())

    @property
    def prev_content(self) -> StateEventContent:
        if self.unsigned and self.unsigned.prev_content:
            return self.unsigned.prev_content
        return state_event_content_map.get(self.type, Obj)()

    @classmethod
    def deserialize(cls, data: JSON) -> "StateEvent":
        try:
            event_type = EventType.find(data.get("type"), t_class=EventType.Class.STATE)
            data.get("content", {})["__mautrix_event_type"] = event_type
            if "prev_content" in data and "prev_content" not in (data.get("unsigned") or {}):
                # This if is a workaround for Conduit being extremely dumb
                if data.get("unsigned", {}) is None:
                    data["unsigned"] = {}
                data.setdefault("unsigned", {})["prev_content"] = data["prev_content"]
            data.get("unsigned", {}).get("prev_content", {})["__mautrix_event_type"] = event_type
        except ValueError:
            return Obj(**data)
        evt = super().deserialize(data)
        evt.type = event_type
        return evt

    @staticmethod
    @deserializer(StateEventContent)
    def deserialize_content(data: JSON) -> StateEventContent:
        evt_type = data.pop("__mautrix_event_type", None)
        content_type = state_event_content_map.get(evt_type, None)
        if not content_type:
            return Obj(**data)
        return content_type.deserialize(data)



# From https://github.com/Lonami/dumbot/blob/master/dumbot.py
# Modified to add Serializable base
from __future__ import annotations

from ..primitive import JSON
from .serializable import AbstractSerializable, Serializable


class Obj(AbstractSerializable):
    """"""

    def __init__(self, **kwargs):
        self.__dict__ = {
            k: Obj(**v) if isinstance(v, dict) else (Lst(v) if isinstance(v, list) else v)
            for k, v in kwargs.items()
        }

    def __getattr__(self, name):
        name = name.rstrip("_")
        obj = self.__dict__.get(name)
        if obj is None:
            obj = Obj()
            self.__dict__[name] = obj
        return obj

    def __getitem__(self, name):
        return self.__dict__.get(name)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __str__(self):
        return str(self.serialize())

    def __repr__(self):
        return repr(self.serialize())

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __bool__(self):
        return bool(self.__dict__)

    def __contains__(self, item):
        return item in self.__dict__

    def popitem(self):
        return self.__dict__.popitem()

    def get(self, key, default=None):
        obj = self.__dict__.get(key)
        if obj is None:
            return default
        else:
            return obj

    def serialize(self) -> dict[str, JSON]:
        return {
            k: v.serialize() if isinstance(v, Serializable) else v
            for k, v in self.__dict__.items()
        }

    @classmethod
    def deserialize(cls, data: dict[str, JSON]) -> Obj:
        return cls(**data)


class Lst(list, AbstractSerializable):
    def __init__(self, iterable=()):
        list.__init__(
            self,
            (
                Obj(**x) if isinstance(x, dict) else (Lst(x) if isinstance(x, list) else x)
                for x in iterable
            ),
        )

    def __repr__(self):
        return super().__repr__()

    def serialize(self) -> list[JSON]:
        return [v.serialize() if isinstance(v, Serializable) else v for v in self]

    @classmethod
    def deserialize(cls, data: list[JSON]) -> Lst:
        return cls(data)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, NewType, Union

from attr import dataclass

from ..primitive import JSON, EventID, RoomID, UserID
from ..util import ExtensibleEnum, SerializableAttrs, SerializableEnum, deserializer
from .base import BaseEvent, GenericEvent
from .type import EventType


@dataclass
class TypingEventContent(SerializableAttrs):
    user_ids: List[UserID]


@dataclass
class TypingEvent(BaseEvent, SerializableAttrs):
    room_id: RoomID
    content: TypingEventContent


class PresenceState(SerializableEnum):
    ONLINE = "online"
    OFFLINE = "offline"
    UNAVAILABLE = "unavailable"


@dataclass
class PresenceEventContent(SerializableAttrs):
    presence: PresenceState
    last_active_ago: int = None
    status_msg: str = None
    currently_active: bool = None


@dataclass
class PresenceEvent(BaseEvent, SerializableAttrs):
    sender: UserID
    content: PresenceEventContent


@dataclass
class SingleReceiptEventContent(SerializableAttrs):
    ts: int


class ReceiptType(ExtensibleEnum):
    READ = "m.read"
    READ_PRIVATE = "m.read.private"


ReceiptEventContent = Dict[EventID, Dict[ReceiptType, Dict[UserID, SingleReceiptEventContent]]]


@dataclass
class ReceiptEvent(BaseEvent, SerializableAttrs):
    room_id: RoomID
    content: ReceiptEventContent


EphemeralEvent = NewType("EphemeralEvent", Union[PresenceEvent, TypingEvent, ReceiptEvent])


@deserializer(EphemeralEvent)
def deserialize_ephemeral_event(data: JSON) -> EphemeralEvent:
    event_type = EventType.find(data.get("type", None))
    if event_type == EventType.RECEIPT:
        evt = ReceiptEvent.deserialize(data)
    elif event_type == EventType.TYPING:
        evt = TypingEvent.deserialize(data)
    elif event_type == EventType.PRESENCE:
        evt = PresenceEvent.deserialize(data)
    else:
        evt = GenericEvent.deserialize(data)
    evt.type = event_type
    return evt


setattr(EphemeralEvent, "deserialize", deserialize_ephemeral_event)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from mautrix.types import EventID

from .handler import (
    SECTION_GENERAL,
    CommandEvent,
    HelpCacheKey,
    HelpSection,
    command_handler,
    command_handlers,
)


@command_handler(
    needs_auth=False, help_section=SECTION_GENERAL, help_text="Cancel an ongoing action."
)
async def cancel(evt: CommandEvent) -> EventID:
    if evt.sender.command_status:
        action = evt.sender.command_status["action"]
        evt.sender.command_status = None
        return await evt.reply(f"{action} cancelled.")
    else:
        return await evt.reply("No ongoing command.")


@command_handler(
    needs_auth=False, help_section=SECTION_GENERAL, help_text="Get the bridge version."
)
async def version(evt: CommandEvent) -> None:
    if not evt.processor.bridge:
        await evt.reply("Bridge version unknown")
    else:
        await evt.reply(
            f"[{evt.processor.bridge.name}]({evt.processor.bridge.repo_url}) "
            f"{evt.processor.bridge.markdown_version or evt.processor.bridge.version}"
        )


@command_handler(needs_auth=False)
async def unknown_command(evt: CommandEvent) -> EventID:
    return await evt.reply("Unknown command. Try `$cmdprefix+sp help` for help.")


help_cache: dict[HelpCacheKey, str] = {}


async def _get_help_text(evt: CommandEvent) -> str:
    cache_key = await evt.get_help_key()
    if cache_key not in help_cache:
        help_sections: dict[HelpSection, list[str]] = {}
        for handler in command_handlers.values():
            if (
                handler.has_help
                and handler.has_permission(cache_key)
                and handler.is_enabled_for(evt)
            ):
                help_sections.setdefault(handler.help_section, [])
                help_sections[handler.help_section].append(handler.help + "  ")
        help_sorted = sorted(help_sections.items(), key=lambda item: item[0].order)
        helps = ["#### {}\n{}\n".format(key.name, "\n".join(value)) for key, value in help_sorted]
        help_cache[cache_key] = "\n".join(helps)
    return help_cache[cache_key]


def _get_management_status(evt: CommandEvent) -> str:
    if evt.is_management:
        return "This is a management room: prefixing commands with `$cmdprefix` is not required."
    elif evt.is_portal:
        return (
            "**This is a portal room**: you must always prefix commands with `$cmdprefix`.\n"
            "Management commands will not be bridged."
        )
    return "**This is not a management room**: you must prefix commands with `$cmdprefix`."


@command_handler(
    name="help",
    needs_auth=False,
    help_section=SECTION_GENERAL,
    help_text="Show this help message.",
)
async def help_cmd(evt: CommandEvent) -> EventID:
    return await evt.reply(_get_management_status(evt) + "\n" + await _get_help_text(evt))



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Awaitable
from datetime import datetime, timezone
import asyncio

from aiohttp import ClientSession
from yarl import URL

from mautrix.api import HTTPAPI, Method, PathBuilder
from mautrix.types import UserID
from mautrix.util.logging import TraceLogger

from .. import api as as_api, state_store as ss


class AppServiceAPI(HTTPAPI):
    """
    AppServiceAPI is an extension to HTTPAPI that provides appservice-specific features,
    such as child instances and easy access to IntentAPIs.
    """

    base_log: TraceLogger

    identity: UserID | None
    bot_mxid: UserID

    state_store: ss.ASStateStore
    txn_id: int
    children: dict[str, ChildAppServiceAPI]
    real_users: dict[str, AppServiceAPI]

    is_real_user: bool
    bridge_name: str | None

    _bot_intent: as_api.IntentAPI | None

    def __init__(
        self,
        base_url: URL | str,
        bot_mxid: UserID = None,
        token: str = None,
        identity: UserID | None = None,
        log: TraceLogger = None,
        state_store: ss.ASStateStore = None,
        client_session: ClientSession = None,
        child: bool = False,
        real_user: bool = False,
        bridge_name: str | None = None,
        default_retry_count: int = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        """
        Args:
            base_url: The base URL of the homeserver client-server API to use.
            bot_mxid: The Matrix user ID of the appservice bot.
            token: The access token to use.
            identity: The ID of the Matrix user to act as.
            log: The logging.Logger instance to log requests with.
            state_store: The StateStore instance to use.
            client_session: The aiohttp ClientSession to use.
            child: Whether or not this is instance is a child of another AppServiceAPI.
            real_user: Whether or not this is a real (non-appservice-managed) user.
            bridge_name: The name of the bridge to put in the ``fi.mau.double_puppet_source`` field
                in outgoing message events sent through real users.
        """
        self.base_log = log
        api_log = self.base_log.getChild("api").getChild(identity or "bot")
        super().__init__(
            base_url=base_url,
            token=token,
            loop=loop,
            log=api_log,
            client_session=client_session,
            txn_id=0 if not child else None,
            default_retry_count=default_retry_count,
        )
        self.identity = identity
        self.bot_mxid = bot_mxid
        self._bot_intent = None
        self.state_store = state_store
        self.is_real_user = real_user
        self.bridge_name = bridge_name

        if not child:
            self.txn_id = 0
            if not real_user:
                self.children = {}
                self.real_users = {}

    def user(self, user: UserID) -> ChildAppServiceAPI:
        """
        Get the AppServiceAPI for an appservice-managed user.

        Args:
            user: The Matrix user ID of the user whose AppServiceAPI to get.

        Returns:
            The ChildAppServiceAPI object for the user.
        """
        if self.is_real_user:
            raise ValueError("Can't get child of real user")

        try:
            return self.children[user]
        except KeyError:
            child = ChildAppServiceAPI(user, self)
            self.children[user] = child
            return child

    def real_user(self, mxid: UserID, token: str, base_url: URL | None = None) -> AppServiceAPI:
        """
        Get the AppServiceAPI for a real (non-appservice-managed) Matrix user.

        Args:
            mxid: The Matrix user ID of the user whose AppServiceAPI to get.
            token: The access token for the user.
            base_url: The base URL of the homeserver client-server API to use. Defaults to the
                appservice homeserver URL.

        Returns:
            The AppServiceAPI object for the user.

        Raises:
            ValueError: When this AppServiceAPI instance is a real user.
        """
        if self.is_real_user:
            raise ValueError("Can't get child of real user")

        try:
            child = self.real_users[mxid]
            child.base_url = base_url or child.base_url
            child.token = token or child.token
        except KeyError:
            child = type(self)(
                base_url=base_url or self.base_url,
                token=token,
                identity=mxid,
                log=self.base_log,
                state_store=self.state_store,
                client_session=self.session,
                real_user=True,
                bridge_name=self.bridge_name,
                default_retry_count=self.default_retry_count,
            )
            self.real_users[mxid] = child
        return child

    def bot_intent(self) -> as_api.IntentAPI:
        """
        Get the intent API for the appservice bot.

        Returns:
            The IntentAPI object for the appservice bot
        """
        if not self._bot_intent:
            self._bot_intent = as_api.IntentAPI(self.bot_mxid, self, state_store=self.state_store)
        return self._bot_intent

    def intent(
        self, user: UserID = None, token: str | None = None, base_url: str | None = None
    ) -> as_api.IntentAPI:
        """
        Get the intent API of a child user.

        Args:
            user: The Matrix user ID whose intent API to get.
            token: The access token to use. Only applicable for non-appservice-managed users.
            base_url: The base URL of the homeserver client-server API to use. Only applicable for
                non-appservice users. Defaults to the appservice homeserver URL.

        Returns:
            The IntentAPI object for the given user.

        Raises:
            ValueError: When this AppServiceAPI instance is a real user.
        """
        if self.is_real_user:
            raise ValueError("Can't get child intent of real user")
        if token:
            return as_api.IntentAPI(
                user, self.real_user(user, token, base_url), self.bot_intent(), self.state_store
            )
        return as_api.IntentAPI(user, self.user(user), self.bot_intent(), self.state_store)

    def request(
        self,
        method: Method,
        path: PathBuilder,
        content: dict | bytes | str | None = None,
        timestamp: int | None = None,
        headers: dict[str, str] | None = None,
        query_params: dict[str, Any] | None = None,
        retry_count: int | None = None,
        metrics_method: str | None = "",
        min_iter_size: int = 25 * 1024 * 1024,
    ) -> Awaitable[dict]:
        """
        Make a raw Matrix API request, acting as the appservice user assigned to this AppServiceAPI
        instance and optionally including timestamp massaging.

        Args:
            method: The HTTP method to use.
            path: The full API endpoint to call (including the _matrix/... prefix)
            content: The content to post as a dict/list (will be serialized as JSON)
                     or bytes/str (will be sent as-is).
            timestamp: The timestamp query param used for timestamp massaging.
            headers: A dict of HTTP headers to send. If the headers don't contain ``Content-Type``,
                     it'll be set to ``application/json``. The ``Authorization`` header is always
                     overridden if :attr:`token` is set.
            query_params: A dict of query parameters to send.
            retry_count: Number of times to retry if the homeserver isn't reachable.
                         Defaults to :attr:`default_retry_count`.
            metrics_method: Name of the method to include in Prometheus timing metrics.
            min_iter_size: If the request body is larger than this value, it will be passed to
                           aiohttp as an async iterable to stop it from copying the whole thing
                           in memory.

        Returns:
            The parsed response JSON.
        """
        query_params = query_params or {}
        if timestamp is not None:
            if isinstance(timestamp, datetime):
                timestamp = int(timestamp.replace(tzinfo=timezone.utc).timestamp() * 1000)
            query_params["ts"] = timestamp
        if not self.is_real_user:
            query_params["user_id"] = self.identity or self.bot_mxid

        return super().request(
            method, path, content, headers, query_params, retry_count, metrics_method
        )


class ChildAppServiceAPI(AppServiceAPI):
    """
    ChildAppServiceAPI is a simple way to copy AppServiceAPIs while maintaining a shared txn_id.
    """

    parent: AppServiceAPI

    def __init__(self, user: UserID, parent: AppServiceAPI) -> None:
        """
        Args:
            user: The Matrix user ID of the child user.
            parent: The parent AppServiceAPI instance.
        """
        super().__init__(
            parent.base_url,
            parent.bot_mxid,
            parent.token,
            user,
            parent.base_log,
            parent.state_store,
            parent.session,
            child=True,
            bridge_name=parent.bridge_name,
            default_retry_count=parent.default_retry_count,
        )
        self.parent = parent

    @property
    def txn_id(self) -> int:
        return self.parent.txn_id

    @txn_id.setter
    def txn_id(self, value: int) -> None:
        self.parent.txn_id = value



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from collections import defaultdict
from datetime import timedelta

from asyncpg import UniqueViolationError

from mautrix.client.state_store import SyncStore
from mautrix.client.state_store.asyncpg import PgStateStore
from mautrix.errors import GroupSessionWithheldError
from mautrix.types import (
    CrossSigner,
    CrossSigningUsage,
    DeviceID,
    DeviceIdentity,
    EventID,
    IdentityKey,
    RoomID,
    RoomKeyWithheldCode,
    SessionID,
    SigningKey,
    SyncToken,
    TOFUSigningKey,
    TrustState,
    UserID,
)
from mautrix.util.async_db import Database, Scheme
from mautrix.util.logging import TraceLogger

from ... import InboundGroupSession, OlmAccount, OutboundGroupSession, RatchetSafety, Session
from ..abstract import CryptoStore, StateStore
from .upgrade import upgrade_table

try:
    from sqlite3 import IntegrityError, sqlite_version_info as sqlite_version

    from aiosqlite import Cursor
except ImportError:
    Cursor = None
    sqlite_version = (0, 0, 0)

    class IntegrityError(Exception):
        pass


class PgCryptoStateStore(PgStateStore, StateStore):
    """
    This class ensures that the PgStateStore in the client module implements the StateStore
    methods needed by the crypto module.
    """


class PgCryptoStore(CryptoStore, SyncStore):
    upgrade_table = upgrade_table

    db: Database
    account_id: str
    pickle_key: str
    log: TraceLogger

    _sync_token: SyncToken | None
    _device_id: DeviceID | None
    _account: OlmAccount | None
    _olm_cache: dict[IdentityKey, dict[SessionID, Session]]

    def __init__(self, account_id: str, pickle_key: str, db: Database) -> None:
        self.db = db
        self.account_id = account_id
        self.pickle_key = pickle_key
        self.log = db.log

        self._sync_token = None
        self._device_id = DeviceID("")
        self._account = None
        self._olm_cache = defaultdict(lambda: {})

    async def delete(self) -> None:
        tables = ("crypto_account", "crypto_olm_session", "crypto_megolm_outbound_session")
        async with self.db.acquire() as conn, conn.transaction():
            for table in tables:
                await conn.execute(f"DELETE FROM {table} WHERE account_id=$1", self.account_id)

    async def get_device_id(self) -> DeviceID | None:
        q = "SELECT device_id FROM crypto_account WHERE account_id=$1"
        device_id = await self.db.fetchval(q, self.account_id)
        self._device_id = device_id or self._device_id
        return self._device_id

    async def put_device_id(self, device_id: DeviceID) -> None:
        q = "UPDATE crypto_account SET device_id=$1 WHERE account_id=$2"
        await self.db.fetchval(q, device_id, self.account_id)
        self._device_id = device_id

    async def put_next_batch(self, next_batch: SyncToken) -> None:
        self._sync_token = next_batch
        q = "UPDATE crypto_account SET sync_token=$1 WHERE account_id=$2"
        await self.db.execute(q, self._sync_token, self.account_id)

    async def get_next_batch(self) -> SyncToken:
        if self._sync_token is None:
            q = "SELECT sync_token FROM crypto_account WHERE account_id=$1"
            self._sync_token = await self.db.fetchval(q, self.account_id)
        return self._sync_token

    async def put_account(self, account: OlmAccount) -> None:
        self._account = account
        pickle = account.pickle(self.pickle_key)
        q = """
        INSERT INTO crypto_account (account_id, device_id, shared, sync_token, account)
        VALUES ($1, $2, $3, $4, $5)
        ON CONFLICT (account_id) DO UPDATE
            SET shared=excluded.shared, sync_token=excluded.sync_token, account=excluded.account
        """
        await self.db.execute(
            q,
            self.account_id,
            self._device_id or "",
            account.shared,
            self._sync_token or "",
            pickle,
        )

    async def get_account(self) -> OlmAccount:
        if self._account is None:
            q = "SELECT shared, account, device_id FROM crypto_account WHERE account_id=$1"
            row = await self.db.fetchrow(q, self.account_id)
            if row is not None:
                self._account = OlmAccount.from_pickle(
                    row["account"], passphrase=self.pickle_key, shared=row["shared"]
                )
        return self._account

    async def has_session(self, key: IdentityKey) -> bool:
        if len(self._olm_cache[key]) > 0:
            return True
        q = "SELECT session_id FROM crypto_olm_session WHERE sender_key=$1 AND account_id=$2"
        val = await self.db.fetchval(q, key, self.account_id)
        return val is not None

    async def get_sessions(self, key: IdentityKey) -> list[Session]:
        q = """
        SELECT session_id, session, created_at, last_encrypted, last_decrypted
        FROM crypto_olm_session WHERE sender_key=$1 AND account_id=$2
        ORDER BY last_decrypted DESC
        """
        rows = await self.db.fetch(q, key, self.account_id)
        sessions = []
        for row in rows:
            try:
                sess = self._olm_cache[key][row["session_id"]]
            except KeyError:
                sess = Session.from_pickle(
                    row["session"],
                    passphrase=self.pickle_key,
                    creation_time=row["created_at"],
                    last_encrypted=row["last_encrypted"],
                    last_decrypted=row["last_decrypted"],
                )
                self._olm_cache[key][SessionID(sess.id)] = sess
            sessions.append(sess)
        return sessions

    async def get_latest_session(self, key: IdentityKey) -> Session | None:
        q = """
        SELECT session_id, session, created_at, last_encrypted, last_decrypted
        FROM crypto_olm_session WHERE sender_key=$1 AND account_id=$2
        ORDER BY last_decrypted DESC LIMIT 1
        """
        row = await self.db.fetchrow(q, key, self.account_id)
        if row is None:
            return None
        try:
            return self._olm_cache[key][row["session_id"]]
        except KeyError:
            sess = Session.from_pickle(
                row["session"],
                passphrase=self.pickle_key,
                creation_time=row["created_at"],
                last_encrypted=row["last_encrypted"],
                last_decrypted=row["last_decrypted"],
            )
            self._olm_cache[key][SessionID(sess.id)] = sess
            return sess

    async def add_session(self, key: IdentityKey, session: Session) -> None:
        if session.id in self._olm_cache[key]:
            self.log.warning(f"Cache already contains Olm session with ID {session.id}")
        self._olm_cache[key][SessionID(session.id)] = session
        pickle = session.pickle(self.pickle_key)
        q = """
        INSERT INTO crypto_olm_session (
            session_id, sender_key, session, created_at, last_encrypted, last_decrypted, account_id
        ) VALUES ($1, $2, $3, $4, $5, $6, $7)
        """
        await self.db.execute(
            q,
            session.id,
            key,
            pickle,
            session.creation_time,
            session.last_encrypted,
            session.last_decrypted,
            self.account_id,
        )

    async def update_session(self, key: IdentityKey, session: Session) -> None:
        try:
            assert self._olm_cache[key][SessionID(session.id)] == session
        except (KeyError, AssertionError) as e:
            self.log.warning(
                f"Cached olm session with ID {session.id} "
                f"isn't equal to the one being saved to the database ({e})"
            )
        pickle = session.pickle(self.pickle_key)
        q = """
        UPDATE crypto_olm_session SET session=$1, last_encrypted=$2, last_decrypted=$3
        WHERE session_id=$4 AND account_id=$5
        """
        await self.db.execute(
            q, pickle, session.last_encrypted, session.last_decrypted, session.id, self.account_id
        )

    async def put_group_session(
        self,
        room_id: RoomID,
        sender_key: IdentityKey,
        session_id: SessionID,
        session: InboundGroupSession,
    ) -> None:
        pickle = session.pickle(self.pickle_key)
        forwarding_chains = ",".join(session.forwarding_chain)
        q = """
        INSERT INTO crypto_megolm_inbound_session (
            session_id, sender_key, signing_key, room_id, session, forwarding_chains,
            ratchet_safety, received_at, max_age, max_messages, is_scheduled, account_id
        ) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12)
        ON CONFLICT (session_id, account_id) DO UPDATE
        SET withheld_code=NULL, withheld_reason=NULL, sender_key=excluded.sender_key,
            signing_key=excluded.signing_key, room_id=excluded.room_id, session=excluded.session,
            forwarding_chains=excluded.forwarding_chains, ratchet_safety=excluded.ratchet_safety,
            received_at=excluded.received_at, max_age=excluded.max_age,
            max_messages=excluded.max_messages, is_scheduled=excluded.is_scheduled
        """
        try:
            await self.db.execute(
                q,
                session_id,
                sender_key,
                session.signing_key,
                room_id,
                pickle,
                forwarding_chains,
                session.ratchet_safety.json(),
                session.received_at,
                int(session.max_age.total_seconds() * 1000) if session.max_age else None,
                session.max_messages,
                session.is_scheduled,
                self.account_id,
            )
        except (IntegrityError, UniqueViolationError):
            self.log.exception(f"Failed to insert megolm session {session_id}")

    async def get_group_session(
        self, room_id: RoomID, session_id: SessionID
    ) -> InboundGroupSession | None:
        q = """
        SELECT
            sender_key, signing_key, session, forwarding_chains, withheld_code,
            ratchet_safety, received_at, max_age, max_messages, is_scheduled
        FROM crypto_megolm_inbound_session
        WHERE room_id=$1 AND session_id=$2 AND account_id=$3
        """
        row = await self.db.fetchrow(q, room_id, session_id, self.account_id)
        if row is None:
            return None
        if row["withheld_code"] is not None:
            raise GroupSessionWithheldError(session_id, row["withheld_code"])
        forwarding_chain = row["forwarding_chains"].split(",") if row["forwarding_chains"] else []
        return InboundGroupSession.from_pickle(
            row["session"],
            passphrase=self.pickle_key,
            signing_key=row["signing_key"],
            sender_key=row["sender_key"],
            room_id=room_id,
            forwarding_chain=forwarding_chain,
            ratchet_safety=RatchetSafety.parse_json(row["ratchet_safety"] or "{}"),
            received_at=row["received_at"],
            max_age=timedelta(milliseconds=row["max_age"]) if row["max_age"] else None,
            max_messages=row["max_messages"],
            is_scheduled=row["is_scheduled"],
        )

    async def redact_group_session(
        self, room_id: RoomID, session_id: SessionID, reason: str
    ) -> None:
        q = """
        UPDATE crypto_megolm_inbound_session
        SET withheld_code=$1, withheld_reason=$2, session=NULL, forwarding_chains=NULL
        WHERE session_id=$3 AND account_id=$4 AND session IS NOT NULL
        """
        await self.db.execute(
            q,
            RoomKeyWithheldCode.BEEPER_REDACTED.value,
            f"Session redacted: {reason}",
            session_id,
            self.account_id,
        )

    async def redact_group_sessions(
        self, room_id: RoomID, sender_key: IdentityKey, reason: str
    ) -> list[SessionID]:
        if not room_id and not sender_key:
            raise ValueError("Either room_id or sender_key must be provided")
        q = """
        UPDATE crypto_megolm_inbound_session
        SET withheld_code=$1, withheld_reason=$2, session=NULL, forwarding_chains=NULL
        WHERE (room_id=$3 OR $3='') AND (sender_key=$4 OR $4='') AND account_id=$5
          AND session IS NOT NULL AND is_scheduled=false AND received_at IS NOT NULL
        RETURNING session_id
        """
        rows = await self.db.fetch(
            q,
            RoomKeyWithheldCode.BEEPER_REDACTED.value,
            f"Session redacted: {reason}",
            room_id,
            sender_key,
            self.account_id,
        )
        return [row["session_id"] for row in rows]

    async def redact_expired_group_sessions(self) -> list[SessionID]:
        if self.db.scheme == Scheme.SQLITE:
            q = """
            UPDATE crypto_megolm_inbound_session
            SET withheld_code=$1, withheld_reason=$2, session=NULL, forwarding_chains=NULL
            WHERE account_id=$3 AND session IS NOT NULL AND is_scheduled=false
              AND received_at IS NOT NULL and max_age IS NOT NULL
              AND unixepoch(received_at) + (2 * max_age / 1000) < unixepoch(date('now'))
            RETURNING session_id
            """
        elif self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
            q = """
            UPDATE crypto_megolm_inbound_session
            SET withheld_code=$1, withheld_reason=$2, session=NULL, forwarding_chains=NULL
            WHERE account_id=$3 AND session IS NOT NULL AND is_scheduled=false
              AND received_at IS NOT NULL and max_age IS NOT NULL
              AND received_at + 2 * (max_age * interval '1 millisecond') < now()
            RETURNING session_id
            """
        else:
            raise RuntimeError(f"Unsupported dialect {self.db.scheme}")
        rows = await self.db.fetch(
            q,
            RoomKeyWithheldCode.BEEPER_REDACTED.value,
            f"Session redacted: expired",
            self.account_id,
        )
        return [row["session_id"] for row in rows]

    async def has_group_session(self, room_id: RoomID, session_id: SessionID) -> bool:
        q = """
        SELECT COUNT(session) FROM crypto_megolm_inbound_session
        WHERE room_id=$1 AND session_id=$2 AND account_id=$3 AND session IS NOT NULL
        """
        count = await self.db.fetchval(q, room_id, session_id, self.account_id)
        return count > 0

    async def add_outbound_group_session(self, session: OutboundGroupSession) -> None:
        pickle = session.pickle(self.pickle_key)
        max_age = int(session.max_age.total_seconds() * 1000)
        q = """
        INSERT INTO crypto_megolm_outbound_session (
            room_id, session_id, session, shared, max_messages, message_count,
            max_age, created_at, last_used, account_id
        ) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)
        ON CONFLICT (account_id, room_id) DO UPDATE
        SET session_id=excluded.session_id, session=excluded.session, shared=excluded.shared,
            max_messages=excluded.max_messages, message_count=excluded.message_count,
            max_age=excluded.max_age, created_at=excluded.created_at, last_used=excluded.last_used
        """
        await self.db.execute(
            q,
            session.room_id,
            session.id,
            pickle,
            session.shared,
            session.max_messages,
            session.message_count,
            max_age,
            session.creation_time,
            session.use_time,
            self.account_id,
        )

    async def update_outbound_group_session(self, session: OutboundGroupSession) -> None:
        pickle = session.pickle(self.pickle_key)
        q = """
        UPDATE crypto_megolm_outbound_session SET session=$1, message_count=$2, last_used=$3
        WHERE room_id=$4 AND session_id=$5 AND account_id=$6
        """
        await self.db.execute(
            q,
            pickle,
            session.message_count,
            session.use_time,
            session.room_id,
            session.id,
            self.account_id,
        )

    async def get_outbound_group_session(self, room_id: RoomID) -> OutboundGroupSession | None:
        q = """
        SELECT room_id, session_id, session, shared, max_messages, message_count, max_age,
               created_at, last_used
        FROM crypto_megolm_outbound_session WHERE room_id=$1 AND account_id=$2
        """
        row = await self.db.fetchrow(q, room_id, self.account_id)
        if row is None:
            return None
        return OutboundGroupSession.from_pickle(
            row["session"],
            passphrase=self.pickle_key,
            room_id=row["room_id"],
            shared=row["shared"],
            max_messages=row["max_messages"],
            message_count=row["message_count"],
            max_age=timedelta(milliseconds=row["max_age"]),
            use_time=row["last_used"],
            creation_time=row["created_at"],
        )

    async def remove_outbound_group_session(self, room_id: RoomID) -> None:
        q = "DELETE FROM crypto_megolm_outbound_session WHERE room_id=$1 AND account_id=$2"
        await self.db.execute(q, room_id, self.account_id)

    async def remove_outbound_group_sessions(self, rooms: list[RoomID]) -> None:
        if self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
            q = """
            DELETE FROM crypto_megolm_outbound_session WHERE account_id=$1 AND room_id=ANY($2)
            """
            await self.db.execute(q, self.account_id, rooms)
        else:
            params = ",".join(["?"] * len(rooms))
            q = f"""
            DELETE FROM crypto_megolm_outbound_session WHERE account_id=? AND room_id IN ({params})
            """
            await self.db.execute(q, self.account_id, *rooms)

    _validate_message_index_query = """
    INSERT INTO crypto_message_index (sender_key, session_id, "index", event_id, timestamp)
    VALUES ($1, $2, $3, $4, $5)
    -- have to update something so that RETURNING * always returns the row
    ON CONFLICT (sender_key, session_id, "index") DO UPDATE SET sender_key=excluded.sender_key
    RETURNING *
    """

    async def validate_message_index(
        self,
        sender_key: IdentityKey,
        session_id: SessionID,
        event_id: EventID,
        index: int,
        timestamp: int,
    ) -> bool:
        if self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH) or (
            # RETURNING was added in SQLite 3.35.0 https://www.sqlite.org/lang_returning.html
            self.db.scheme == Scheme.SQLITE
            and sqlite_version >= (3, 35)
        ):
            row = await self.db.fetchrow(
                self._validate_message_index_query,
                sender_key,
                session_id,
                index,
                event_id,
                timestamp,
            )
            return row["event_id"] == event_id and row["timestamp"] == timestamp
        else:
            row = await self.db.fetchrow(
                "SELECT event_id, timestamp FROM crypto_message_index "
                'WHERE sender_key=$1 AND session_id=$2 AND "index"=$3',
                sender_key,
                session_id,
                index,
            )
            if row is not None:
                return row["event_id"] == event_id and row["timestamp"] == timestamp
            await self.db.execute(
                "INSERT INTO crypto_message_index(sender_key, session_id, "
                '                                 "index", event_id, timestamp) '
                "VALUES ($1, $2, $3, $4, $5)",
                sender_key,
                session_id,
                index,
                event_id,
                timestamp,
            )
            return True

    async def get_devices(self, user_id: UserID) -> dict[DeviceID, DeviceIdentity] | None:
        q = "SELECT user_id FROM crypto_tracked_user WHERE user_id=$1"
        tracked_user_id = await self.db.fetchval(q, user_id)
        if tracked_user_id is None:
            return None
        q = """
        SELECT device_id, identity_key, signing_key, trust, deleted, name
        FROM crypto_device WHERE user_id=$1
        """
        rows = await self.db.fetch(q, user_id)
        result = {}
        for row in rows:
            result[row["device_id"]] = DeviceIdentity(
                user_id=user_id,
                device_id=row["device_id"],
                identity_key=row["identity_key"],
                signing_key=row["signing_key"],
                trust=TrustState(row["trust"]),
                deleted=row["deleted"],
                name=row["name"],
            )
        return result

    async def get_device(self, user_id: UserID, device_id: DeviceID) -> DeviceIdentity | None:
        q = """
        SELECT identity_key, signing_key, trust, deleted, name FROM crypto_device
        WHERE user_id=$1 AND device_id=$2
        """
        row = await self.db.fetchrow(q, user_id, device_id)
        if row is None:
            return None
        return DeviceIdentity(
            user_id=user_id,
            device_id=device_id,
            name=row["name"],
            identity_key=row["identity_key"],
            signing_key=row["signing_key"],
            trust=TrustState(row["trust"]),
            deleted=row["deleted"],
        )

    async def find_device_by_key(
        self, user_id: UserID, identity_key: IdentityKey
    ) -> DeviceIdentity | None:
        q = """
        SELECT device_id, signing_key, trust, deleted, name FROM crypto_device
        WHERE user_id=$1 AND identity_key=$2
        """
        row = await self.db.fetchrow(
            q,
            user_id,
            identity_key,
        )
        if row is None:
            return None
        return DeviceIdentity(
            user_id=user_id,
            device_id=row["device_id"],
            name=row["name"],
            identity_key=identity_key,
            signing_key=row["signing_key"],
            trust=TrustState(row["trust"]),
            deleted=row["deleted"],
        )

    async def put_devices(self, user_id: UserID, devices: dict[DeviceID, DeviceIdentity]) -> None:
        data = [
            (
                user_id,
                device_id,
                identity.identity_key,
                identity.signing_key,
                identity.trust,
                identity.deleted,
                identity.name,
            )
            for device_id, identity in devices.items()
        ]
        columns = [
            "user_id",
            "device_id",
            "identity_key",
            "signing_key",
            "trust",
            "deleted",
            "name",
        ]
        async with self.db.acquire() as conn, conn.transaction():
            q = """
            INSERT INTO crypto_tracked_user (user_id) VALUES ($1) ON CONFLICT (user_id) DO NOTHING
            """
            await conn.execute(q, user_id)
            await conn.execute("DELETE FROM crypto_device WHERE user_id=$1", user_id)
            if self.db.scheme == Scheme.POSTGRES:
                await conn.copy_records_to_table("crypto_device", records=data, columns=columns)
            else:
                q = """
                INSERT INTO crypto_device (
                    user_id, device_id, identity_key, signing_key, trust, deleted, name
                ) VALUES ($1, $2, $3, $4, $5, $6, $7)
                """
                await conn.executemany(q, data)

    async def filter_tracked_users(self, users: list[UserID]) -> list[UserID]:
        if self.db.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
            q = "SELECT user_id FROM crypto_tracked_user WHERE user_id = ANY($1)"
            rows = await self.db.fetch(q, users)
        else:
            params = ",".join(["?"] * len(users))
            q = f"SELECT user_id FROM crypto_tracked_user WHERE user_id IN ({params})"
            rows = await self.db.fetch(q, *users)
        return [row["user_id"] for row in rows]

    async def put_cross_signing_key(
        self, user_id: UserID, usage: CrossSigningUsage, key: SigningKey
    ) -> None:
        q = """
        INSERT INTO crypto_cross_signing_keys (user_id, usage, key, first_seen_key)
        VALUES ($1, $2, $3, $4)
        ON CONFLICT (user_id, usage) DO UPDATE SET key=excluded.key
        """
        try:
            await self.db.execute(q, user_id, usage.value, key, key)
        except Exception:
            self.log.exception(f"Failed to store cross-signing key {user_id}/{key}/{usage}")

    async def get_cross_signing_keys(
        self, user_id: UserID
    ) -> dict[CrossSigningUsage, TOFUSigningKey]:
        q = "SELECT usage, key, first_seen_key FROM crypto_cross_signing_keys WHERE user_id=$1"
        return {
            CrossSigningUsage(row["usage"]): TOFUSigningKey(
                key=SigningKey(row["key"]),
                first=SigningKey(row["first_seen_key"]),
            )
            for row in await self.db.fetch(q, user_id)
        }

    async def put_signature(
        self, target: CrossSigner, signer: CrossSigner, signature: str
    ) -> None:
        q = """
        INSERT INTO crypto_cross_signing_signatures (
            signed_user_id, signed_key, signer_user_id, signer_key, signature
        ) VALUES ($1, $2, $3, $4, $5)
        ON CONFLICT (signed_user_id, signed_key, signer_user_id, signer_key)
            DO UPDATE SET signature=excluded.signature
        """
        signed_user_id, signed_key = target
        signer_user_id, signer_key = signer
        try:
            await self.db.execute(
                q, signed_user_id, signed_key, signer_user_id, signer_key, signature
            )
        except Exception:
            self.log.exception(
                f"Failed to store signature from {signer_user_id}/{signer_key} "
                f"for {signed_user_id}/{signed_key}"
            )

    async def is_key_signed_by(self, target: CrossSigner, signer: CrossSigner) -> bool:
        q = """
        SELECT EXISTS(
            SELECT 1 FROM crypto_cross_signing_signatures
            WHERE signed_user_id=$1 AND signed_key=$2 AND signer_user_id=$3 AND signer_key=$4
        )
        """
        signed_user_id, signed_key = target
        signer_user_id, signer_key = signer
        return await self.db.fetchval(q, signed_user_id, signed_key, signer_user_id, signer_key)

    async def drop_signatures_by_key(self, signer: CrossSigner) -> int:
        signer_user_id, signer_key = signer
        q = "DELETE FROM crypto_cross_signing_signatures WHERE signer_user_id=$1 AND signer_key=$2"
        try:
            res = await self.db.execute(q, signer_user_id, signer_key)
        except Exception:
            self.log.exception(
                f"Failed to drop old signatures made by replaced key {signer_user_id}/{signer_key}"
            )
            return -1
        if Cursor is not None and isinstance(res, Cursor):
            return res.rowcount
        elif (
            isinstance(res, str)
            and res.startswith("DELETE ")
            and (intPart := res[len("DELETE ") :]).isdecimal()
        ):
            return int(intPart)
        return -1



# Copyright (c) 2023 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Awaitable, Callable, TypedDict
import asyncio
import functools
import json

import olm

from mautrix.errors import MForbidden, MNotFound
from mautrix.types import (
    DeviceID,
    EncryptionKeyAlgorithm,
    EventType,
    IdentityKey,
    KeyID,
    RequestedKeyInfo,
    RoomEncryptionStateEventContent,
    RoomID,
    RoomKeyEventContent,
    SessionID,
    SigningKey,
    TrustState,
    UserID,
)
from mautrix.util.logging import TraceLogger

from .. import client as cli, crypto


class SignedObject(TypedDict):
    signatures: dict[UserID, dict[str, str]]
    unsigned: Any


class BaseOlmMachine:
    client: cli.Client
    log: TraceLogger
    crypto_store: crypto.CryptoStore
    state_store: crypto.StateStore

    account: account.OlmAccount

    send_keys_min_trust: TrustState
    share_keys_min_trust: TrustState
    allow_key_share: Callable[[crypto.DeviceIdentity, RequestedKeyInfo], Awaitable[bool]]

    delete_outbound_keys_on_ack: bool
    dont_store_outbound_keys: bool
    delete_previous_keys_on_receive: bool
    ratchet_keys_on_decrypt: bool
    delete_fully_used_keys_on_decrypt: bool
    delete_keys_on_device_delete: bool
    disable_device_change_key_rotation: bool

    # Futures that wait for responses to a key request
    _key_request_waiters: dict[SessionID, asyncio.Future]
    # Futures that wait for a session to be received (either normally or through a key request)
    _inbound_session_waiters: dict[SessionID, asyncio.Future]

    _prev_unwedge: dict[IdentityKey, float]
    _fetch_keys_lock: asyncio.Lock
    _megolm_decrypt_lock: asyncio.Lock
    _cs_fetch_attempted: set[UserID]

    async def wait_for_session(
        self, room_id: RoomID, session_id: SessionID, timeout: float = 3
    ) -> bool:
        try:
            fut = self._inbound_session_waiters[session_id]
        except KeyError:
            fut = asyncio.get_running_loop().create_future()
            self._inbound_session_waiters[session_id] = fut
        try:
            return await asyncio.wait_for(asyncio.shield(fut), timeout)
        except asyncio.TimeoutError:
            return await self.crypto_store.has_group_session(room_id, session_id)

    def _mark_session_received(self, session_id: SessionID) -> None:
        try:
            self._inbound_session_waiters.pop(session_id).set_result(True)
        except KeyError:
            return

    async def _fill_encryption_info(self, evt: RoomKeyEventContent) -> None:
        encryption_info = await self.state_store.get_encryption_info(evt.room_id)
        if not encryption_info:
            self.log.warning(
                f"Encryption info for {evt.room_id} not found in state store, fetching from server"
            )
            try:
                encryption_info = await self.client.get_state_event(
                    evt.room_id, EventType.ROOM_ENCRYPTION
                )
            except (MNotFound, MForbidden) as e:
                self.log.warning(
                    f"Failed to get encryption info for {evt.room_id} from server: {e},"
                    " using defaults"
                )
                encryption_info = RoomEncryptionStateEventContent()
            if not encryption_info:
                self.log.warning(
                    f"Didn't find encryption info for {evt.room_id} on server either,"
                    " using defaults"
                )
                encryption_info = RoomEncryptionStateEventContent()

        if not evt.beeper_max_age_ms:
            evt.beeper_max_age_ms = encryption_info.rotation_period_ms
        if not evt.beeper_max_messages:
            evt.beeper_max_messages = encryption_info.rotation_period_msgs


canonical_json = functools.partial(
    json.dumps, ensure_ascii=False, separators=(",", ":"), sort_keys=True
)


def verify_signature_json(
    data: "SignedObject", user_id: UserID, key_name: DeviceID | str, key: SigningKey
) -> bool:
    data_copy = {**data}
    data_copy.pop("unsigned", None)
    signatures = data_copy.pop("signatures")
    key_id = str(KeyID(EncryptionKeyAlgorithm.ED25519, key_name))
    try:
        signature = signatures[user_id][key_id]
    except KeyError:
        return False
    signed_data = canonical_json(data_copy)
    try:
        olm.ed25519_verify(key, signed_data, signature)
        return True
    except olm.OlmVerifyError:
        return False



from .api import ClientAPI
from .client import Client
from .dispatcher import Dispatcher, MembershipEventDispatcher, SimpleDispatcher
from .encryption_manager import DecryptionDispatcher, EncryptingAPI
from .state_store import FileStateStore, MemoryStateStore, MemorySyncStore, StateStore, SyncStore
from .store_updater import StoreUpdatingAPI
from .syncer import EventHandler, InternalEventType, Syncer, SyncStream

__all__ = [
    "ClientAPI",
    "Client",
    "Dispatcher",
    "MembershipEventDispatcher",
    "SimpleDispatcher",
    "DecryptionDispatcher",
    "EncryptingAPI",
    "FileStateStore",
    "MemoryStateStore",
    "MemorySyncStore",
    "StateStore",
    "SyncStore",
    "StoreUpdatingAPI",
    "EventHandler",
    "InternalEventType",
    "Syncer",
    "SyncStream",
    "state_store",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import time

from mautrix.types import EventType, IdentityKey, Obj, UserID

from .decrypt_olm import OlmDecryptionMachine
from .device_lists import DeviceListMachine
from .encrypt_olm import OlmEncryptionMachine

MIN_UNWEDGE_INTERVAL = 1 * 60 * 60


class OlmUnwedgingMachine(OlmDecryptionMachine, OlmEncryptionMachine, DeviceListMachine):
    async def _unwedge_session(self, sender: UserID, sender_key: IdentityKey) -> None:
        try:
            prev_unwedge = self._prev_unwedge[sender_key]
        except KeyError:
            pass
        else:
            delta = time.monotonic() - prev_unwedge
            if delta < MIN_UNWEDGE_INTERVAL:
                self.log.debug(
                    f"Not creating new Olm session with {sender}/{sender_key}, "
                    f"previous recreation was {delta}s ago"
                )
                return
        self._prev_unwedge[sender_key] = time.monotonic()
        try:
            device = await self.get_or_fetch_device_by_key(sender, sender_key)
            if device is None:
                self.log.warning(
                    f"Didn't find identity of {sender}/{sender_key}, can't unwedge session"
                )
                return
            self.log.debug(
                f"Creating new Olm session with {sender}/{device.user_id} (key: {sender_key})"
            )
            await self.send_encrypted_to_device(
                device, EventType.TO_DEVICE_DUMMY, Obj(), _force_recreate_session=True
            )
        except Exception:
            self.log.exception(f"Error unwedging session with {sender}/{sender_key}")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Callable, Generic, Iterable, TypeVar

T = TypeVar("T")
Number = int | float

class Metric:
    name: str
    documentation: str
    unit: str
    typ: str
    samples: list[Any]
    def add_sample(
        self,
        name: str,
        labels: Iterable[str],
        value: Any,
        timestamp: Any = None,
        exemplar: Any = None,
    ) -> None: ...

class MetricWrapperBase(Generic[T]):
    def __init__(
        self,
        name: str,
        documentation: str,
        labelnames: Iterable[str] = (),
        namespace: str = "",
        subsystem: str = "",
        unit: str = "",
        registry: Any = None,
        labelvalues: Any = None,
    ) -> None: ...
    def describe(self) -> list[Metric]: ...
    def collect(self) -> list[Metric]: ...
    def labels(self, *labelvalues, **labelkwargs) -> T: ...
    def remove(self, *labelvalues) -> None: ...

class ContextManager:
    def __enter__(self) -> None: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...
    def __call__(self, f) -> None: ...

class Counter(MetricWrapperBase[Counter]):
    def inc(self, amount: Number = 1) -> None: ...
    def count_exceptions(self, exception: Exception = Exception) -> ContextManager: ...

class Gauge(MetricWrapperBase[Gauge]):
    def inc(self, amount: Number = 1) -> None: ...
    def dec(self, amount: Number = 1) -> None: ...
    def set(self, value: Number = 1) -> None: ...
    def set_to_current_time(self) -> None: ...
    def track_inprogress(self) -> ContextManager: ...
    def time(self) -> ContextManager: ...
    def set_function(self, f: Callable[[], Number]) -> None: ...

class Summary(MetricWrapperBase[Summary]):
    def observe(self, amount: Number) -> None: ...
    def time(self) -> ContextManager: ...

class Histogram(MetricWrapperBase[Histogram]):
    def __init__(
        self,
        name: str,
        documentation: str,
        labelnames: Iterable[str] = (),
        namespace: str = "",
        subsystem: str = "",
        unit: str = "",
        registry: Any = None,
        labelvalues: Any = None,
        buckets: Iterable[Number] = (),
    ) -> None: ...
    def observe(self, amount: Number = 1) -> None: ...
    def time(self) -> ContextManager: ...

class Info(MetricWrapperBase[Info]):
    def info(self, val: dict[str, str]) -> None: ...

class Enum(MetricWrapperBase[Enum]):
    def __init__(
        self,
        name: str,
        documentation: str,
        labelnames: Iterable[str] = (),
        namespace: str = "",
        subsystem: str = "",
        unit: str = "",
        registry: Any = None,
        labelvalues: Any = None,
        states: Iterable[str] = None,
    ) -> None: ...
    def state(self, state: str) -> None: ...

def async_time(metric: Gauge | Summary | Histogram) -> Callable[[Callable], Callable]: ...



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, ClassVar, Dict, Optional
import logging
import time

from attr import dataclass
import aiohttp

from mautrix.api import HTTPAPI
from mautrix.types import SerializableAttrs, SerializableEnum, UserID, field


class BridgeStateEvent(SerializableEnum):
    #####################################
    # Global state events, no remote ID #
    #####################################

    # Bridge process is starting up
    STARTING = "STARTING"
    # Bridge has started but has no valid credentials
    UNCONFIGURED = "UNCONFIGURED"
    # Bridge is running
    RUNNING = "RUNNING"
    # The server was unable to reach the bridge
    BRIDGE_UNREACHABLE = "BRIDGE_UNREACHABLE"

    ################################################
    # Remote state events, should have a remote ID #
    ################################################

    # Bridge has credentials and has started connecting to a remote network
    CONNECTING = "CONNECTING"
    # Bridge has begun backfilling
    BACKFILLING = "BACKFILLING"
    # Bridge has happily connected and is bridging messages
    CONNECTED = "CONNECTED"
    # Bridge has temporarily disconnected, expected to reconnect automatically
    TRANSIENT_DISCONNECT = "TRANSIENT_DISCONNECT"
    # Bridge has disconnected, will require user to log in again
    BAD_CREDENTIALS = "BAD_CREDENTIALS"
    # Bridge has disconnected for an unknown/unexpected reason - we should investigate
    UNKNOWN_ERROR = "UNKNOWN_ERROR"
    # User has logged out - stop tracking this remote
    LOGGED_OUT = "LOGGED_OUT"


ok_ish_states = (
    BridgeStateEvent.STARTING,
    BridgeStateEvent.UNCONFIGURED,
    BridgeStateEvent.RUNNING,
    BridgeStateEvent.CONNECTING,
    BridgeStateEvent.CONNECTED,
    BridgeStateEvent.BACKFILLING,
)


@dataclass(kw_only=True)
class BridgeState(SerializableAttrs):
    human_readable_errors: ClassVar[Dict[Optional[str], str]] = {}
    default_source: ClassVar[str] = "bridge"
    default_error_ttl: ClassVar[int] = 3600
    default_ok_ttl: ClassVar[int] = 21600

    state_event: BridgeStateEvent
    user_id: Optional[UserID] = None
    remote_id: Optional[str] = None
    remote_name: Optional[str] = None
    timestamp: Optional[int] = None
    ttl: int = 0
    source: Optional[str] = None
    error: Optional[str] = None
    message: Optional[str] = None
    info: Optional[Dict[str, Any]] = None
    reason: Optional[str] = None

    send_attempts_: int = field(default=0, hidden=True)

    def fill(self) -> "BridgeState":
        self.timestamp = self.timestamp or int(time.time())
        self.source = self.source or self.default_source
        if not self.ttl:
            self.ttl = (
                self.default_ok_ttl
                if self.state_event in ok_ish_states
                else self.default_error_ttl
            )
        if self.error:
            try:
                msg = self.human_readable_errors[self.error]
            except KeyError:
                pass
            else:
                self.message = msg.format(message=self.message) if self.message else msg
        return self

    def should_deduplicate(self, prev_state: Optional["BridgeState"]) -> bool:
        if (
            not prev_state
            or prev_state.state_event != self.state_event
            or prev_state.error != self.error
            or prev_state.info != self.info
        ):
            # If there's no previous state or the state was different, send this one.
            return False
        # If the previous state is recent, drop this one
        return prev_state.timestamp + prev_state.ttl > self.timestamp

    async def send(self, url: str, token: str, log: logging.Logger, log_sent: bool = True) -> bool:
        if not url:
            return True
        self.send_attempts_ += 1
        headers = {"Authorization": f"Bearer {token}", "User-Agent": HTTPAPI.default_ua}
        try:
            async with aiohttp.ClientSession() as sess, sess.post(
                url, json=self.serialize(), headers=headers
            ) as resp:
                if not 200 <= resp.status < 300:
                    text = await resp.text()
                    text = text.replace("\n", "\\n")
                    log.warning(
                        f"Unexpected status code {resp.status} "
                        f"sending bridge state update: {text}"
                    )
                    return False
                elif log_sent:
                    log.debug(f"Sent new bridge state {self}")
        except Exception as e:
            log.warning(f"Failed to send updated bridge state: {e}")
            return False
        return True


@dataclass(kw_only=True)
class GlobalBridgeState(SerializableAttrs):
    remote_states: Optional[Dict[str, BridgeState]] = field(json="remoteState", default=None)
    bridge_state: BridgeState = field(json="bridgeState")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Awaitable, Callable, Optional, cast
import functools
import inspect
import logging

from mautrix.util.logging import TraceLogger

from .. import async_db
from .connection import LoggingConnection
from .errors import UnsupportedDatabaseVersion
from .scheme import Scheme

Upgrade = Callable[[LoggingConnection, Scheme], Awaitable[Optional[int]]]
UpgradeWithoutScheme = Callable[[LoggingConnection], Awaitable[Optional[int]]]


async def noop_upgrade(_: LoggingConnection, _2: Scheme) -> None:
    pass


def _wrap_upgrade(fn: UpgradeWithoutScheme | Upgrade) -> Upgrade:
    params = inspect.signature(fn).parameters
    if len(params) == 1:
        _wrapped: UpgradeWithoutScheme = cast(UpgradeWithoutScheme, fn)

        @functools.wraps(_wrapped)
        async def _wrapper(conn: LoggingConnection, _: Scheme) -> Optional[int]:
            return await _wrapped(conn)

        return _wrapper
    else:
        return fn


class UpgradeTable:
    upgrades: list[Upgrade]
    allow_unsupported: bool
    database_name: str
    version_table_name: str
    log: TraceLogger

    def __init__(
        self,
        allow_unsupported: bool = False,
        version_table_name: str = "version",
        database_name: str = "database",
        log: logging.Logger | TraceLogger | None = None,
    ) -> None:
        self.upgrades = []
        self.allow_unsupported = allow_unsupported
        self.version_table_name = version_table_name
        self.database_name = database_name
        self.log = log or logging.getLogger("mau.db.upgrade")

    def register(
        self,
        _outer_fn: Upgrade | UpgradeWithoutScheme | None = None,
        *,
        index: int = -1,
        description: str = "",
        transaction: bool = True,
        upgrades_to: int | Upgrade | None = None,
    ) -> Upgrade | Callable[[Upgrade | UpgradeWithoutScheme], Upgrade]:
        if isinstance(index, str):
            description = index
            index = -1

        def actually_register(fn: Upgrade | UpgradeWithoutScheme) -> Upgrade:
            fn = _wrap_upgrade(fn)
            fn.__mau_db_upgrade_description__ = description
            fn.__mau_db_upgrade_transaction__ = transaction
            fn.__mau_db_upgrade_destination__ = (
                upgrades_to
                if not upgrades_to or isinstance(upgrades_to, int)
                else _wrap_upgrade(upgrades_to)
            )
            if index == -1 or index == len(self.upgrades):
                self.upgrades.append(fn)
            else:
                if len(self.upgrades) <= index:
                    self.upgrades += [noop_upgrade] * (index - len(self.upgrades) + 1)
                self.upgrades[index] = fn
            return fn

        return actually_register(_outer_fn) if _outer_fn else actually_register

    async def _save_version(self, conn: LoggingConnection, version: int) -> None:
        self.log.trace(f"Saving current version (v{version}) to database")
        await conn.execute(f"DELETE FROM {self.version_table_name}")
        await conn.execute(f"INSERT INTO {self.version_table_name} (version) VALUES ($1)", version)

    async def upgrade(self, db: async_db.Database) -> None:
        await db.execute(
            f"""CREATE TABLE IF NOT EXISTS {self.version_table_name} (
                version INTEGER PRIMARY KEY
            )"""
        )
        row = await db.fetchrow(f"SELECT version FROM {self.version_table_name} LIMIT 1")
        version = row["version"] if row else 0

        if len(self.upgrades) < version:
            unsupported_version_error = UnsupportedDatabaseVersion(
                self.database_name, version, len(self.upgrades)
            )
            if not self.allow_unsupported:
                raise unsupported_version_error
            else:
                self.log.warning(str(unsupported_version_error))
                return
        elif len(self.upgrades) == version:
            self.log.debug(f"Database at v{version}, not upgrading")
            return

        async with db.acquire() as conn:
            while version < len(self.upgrades):
                old_version = version
                upgrade = self.upgrades[version]
                new_version = (
                    getattr(upgrade, "__mau_db_upgrade_destination__", None) or version + 1
                )
                if callable(new_version):
                    new_version = await new_version(conn, db.scheme)
                desc = getattr(upgrade, "__mau_db_upgrade_description__", None)
                suffix = f": {desc}" if desc else ""
                self.log.debug(
                    f"Upgrading {self.database_name} from v{old_version} to v{new_version}{suffix}"
                )
                if getattr(upgrade, "__mau_db_upgrade_transaction__", True):
                    async with conn.transaction():
                        version = await upgrade(conn, db.scheme) or new_version
                        await self._save_version(conn, version)
                else:
                    version = await upgrade(conn, db.scheme) or new_version
                    await self._save_version(conn, version)
                if version != new_version:
                    self.log.warning(
                        f"Upgrading {self.database_name} actually went from v{old_version} "
                        f"to v{version}"
                    )


upgrade_tables: dict[str, UpgradeTable] = {}


def register_upgrade_table_parent_module(name: str) -> None:
    upgrade_tables[name] = UpgradeTable()


def _find_upgrade_table(fn: Upgrade) -> UpgradeTable:
    try:
        module = fn.__module__
    except AttributeError as e:
        raise ValueError(
            "Registering upgrades without an UpgradeTable requires the function "
            "to have the __module__ attribute."
        ) from e
    parts = module.split(".")
    used_parts = []
    last_error = None
    for part in parts:
        used_parts.append(part)
        try:
            return upgrade_tables[".".join(used_parts)]
        except KeyError as e:
            last_error = e
    raise KeyError(
        "Registering upgrades without an UpgradeTable requires you to register a parent "
        "module with register_upgrade_table_parent_module first."
    ) from last_error


def register_upgrade(index: int = -1, description: str = "") -> Callable[[Upgrade], Upgrade]:
    def actually_register(fn: Upgrade) -> Upgrade:
        return _find_upgrade_table(fn).register(fn, index=index, description=description)

    return actually_register



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Union

from attr import dataclass
import attr

from ..primitive import JSON, RoomID, UserID
from ..util import Obj, SerializableAttrs, deserializer
from .base import BaseEvent, EventType


@dataclass
class RoomTagInfo(SerializableAttrs):
    order: Union[int, float, str] = None


@dataclass
class RoomTagAccountDataEventContent(SerializableAttrs):
    tags: Dict[str, RoomTagInfo] = attr.ib(default=None, metadata={"json": "tags"})


DirectAccountDataEventContent = Dict[UserID, List[RoomID]]

AccountDataEventContent = Union[RoomTagAccountDataEventContent, DirectAccountDataEventContent, Obj]
account_data_event_content_map = {
    EventType.TAG: RoomTagAccountDataEventContent,
    # m.direct doesn't really need deserializing
    # EventType.DIRECT: DirectAccountDataEventContent,
}


# TODO remaining account data event types


@dataclass
class AccountDataEvent(BaseEvent, SerializableAttrs):
    content: AccountDataEventContent

    @classmethod
    def deserialize(cls, data: JSON) -> "AccountDataEvent":
        try:
            evt_type = EventType.find(data.get("type"))
            data.get("content", {})["__mautrix_event_type"] = evt_type
        except ValueError:
            return Obj(**data)
        evt = super().deserialize(data)
        evt.type = evt_type
        return evt

    @staticmethod
    @deserializer(AccountDataEventContent)
    def deserialize_content(data: JSON) -> AccountDataEventContent:
        evt_type = data.pop("__mautrix_event_type", None)
        content_type = account_data_event_content_map.get(evt_type, None)
        if not content_type:
            return Obj(**data)
        return content_type.deserialize(data)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List

from attr import dataclass

from .event import EventType
from .primitive import RoomID, UserID
from .util import SerializableAttrs, SerializableEnum


class EventFormat(SerializableEnum):
    """
    Federation event format enum, as specified in the `create filter endpoint`_.

    .. _create filter endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3useruseridfilter
    """

    CLIENT = "client"
    FEDERATION = "federation"


@dataclass
class EventFilter(SerializableAttrs):
    """
    Event filter object, as specified in the `create filter endpoint`_.

    .. _create filter endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3useruseridfilter
    """

    limit: int = None
    """The maximum number of events to return."""

    not_senders: List[UserID] = None
    """A list of sender IDs to exclude. If this list is absent then no senders are excluded
    A matching sender will be excluded even if it is listed in the :attr:`senders` filter."""

    not_types: List[EventType] = None
    """A list of event types to exclude. If this list is absent then no event types are excluded.
    A matching type will be excluded even if it is listed in the :attr:`types` filter.
    A ``'*'`` can be used as a wildcard to match any sequence of characters."""

    senders: List[UserID] = None
    """A list of senders IDs to include. If this list is absent then all senders are included."""

    types: List[EventType] = None
    """A list of event types to include. If this list is absent then all event types are included.
    A ``'*'`` can be used as a wildcard to match any sequence of characters."""


@dataclass
class RoomEventFilter(EventFilter, SerializableAttrs):
    """
    Room event filter object, as specified in the `create filter endpoint`_.

    .. _create filter endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3useruseridfilter
    """

    lazy_load_members: bool = False
    """
    If ``True``, enables lazy-loading of membership events. See `Lazy-loading room members`_ for more information.

    .. _Lazy-loading room members:
        https://matrix.org/docs/spec/client_server/r0.5.0#lazy-loading-room-members
    """

    include_redundant_members: bool = False
    """
    If ``True``, sends all membership events for all events, even if they have already been sent
    to the client. Does not apply unless :attr:`lazy_load_members` is true.
    See `Lazy-loading room members`_ for more information."""

    not_rooms: List[RoomID] = None
    """A list of room IDs to exclude. If this list is absent then no rooms are excluded.
    A matching room will be excluded even if it is listed in the :attr:`rooms` filter."""

    rooms: List[RoomID] = None
    """A list of room IDs to include. If this list is absent then all rooms are included."""

    contains_url: bool = None
    """If ``True``, includes only events with a url key in their content. If ``False``, excludes
    those events. If omitted, ``url`` key is not considered for filtering."""


@dataclass
class StateFilter(RoomEventFilter, SerializableAttrs):
    """
    State event filter object, as specified in the `create filter endpoint`_. Currently this is the
    same as :class:`RoomEventFilter`.

    .. _create filter endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3useruseridfilter
    """

    pass


@dataclass
class RoomFilter(SerializableAttrs):
    """
    Room filter object, as specified in the `create filter endpoint`_.

    .. _create filter endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3useruseridfilter
    """

    not_rooms: List[RoomID] = None
    """A list of room IDs to exclude. If this list is absent then no rooms are excluded.
    A matching room will be excluded even if it is listed in the ``'rooms'`` filter.
    This filter is applied before the filters in :attr:`ephemeral`, :attr:`state`,
    :attr:`timeline` or :attr:`account_data`."""

    rooms: List[RoomID] = None
    """A list of room IDs to include. If this list is absent then all rooms are included.
    This filter is applied before the filters in :attr:`ephemeral`, :attr:`state`,
    :attr:`timeline` or :attr:`account_data`."""

    ephemeral: RoomEventFilter = None
    """The events that aren't recorded in the room history, e.g. typing and receipts,
    to include for rooms."""

    include_leave: bool = False
    """Include rooms that the user has left in the sync."""

    state: StateFilter = None
    """The state events to include for rooms."""

    timeline: RoomEventFilter = None
    """The message and state update events to include for rooms."""

    account_data: RoomEventFilter = None
    """The per user account data to include for rooms."""


@dataclass
class Filter(SerializableAttrs):
    """
    Base filter object, as specified in the `create filter endpoint`_.

    .. _create filter endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3useruseridfilter
    """

    event_fields: List[str] = None
    """List of event fields to include. If this list is absent then all fields are included.
    The entries may include ``.`` charaters to indicate sub-fields. So ``['content.body']`` will
    include the ``body`` field of the ``content`` object. A literal ``.`` character in a field name
    may be escaped using a ``\\``. A server may include more fields than were requested."""

    event_format: EventFormat = None
    """The format to use for events. ``'client'`` will return the events in a format suitable for
    clients. ``'federation'`` will return the raw event as receieved over federation. The default
    is :attr:`~EventFormat.CLIENT`."""

    presence: EventFilter = None
    """The presence updates to include."""

    account_data: EventFilter = None
    """The user account data that isn't associated with rooms to include."""

    room: RoomFilter = None
    """Filters to be applied to room data."""
